"""
Contains functions for preprocessing COOM ASP facts and checking user input
"""

from typing import List

from clingo import Control
from clingo.script import enable_python
from clingo.symbol import Symbol

from .utils import get_encoding
from .utils.logging import get_logger

log = get_logger("main")


def preprocess(files: List[str], max_bound: int = 0, discrete: bool = False, multishot: bool = False) -> List[str]:
    """
    Preprocesses COOM ASP facts into a "grounded" configuration fact format
    """
    # pylint: disable=not-context-manager
    input_files = files
    ctl = Control(["-c", f"max_bound={max_bound}"], message_limit=0)
    for f in input_files:
        ctl.load(f)

    enable_python()
    ctl.load(get_encoding("preprocess.lp" if not multishot else "preprocess-multishot.lp"))

    if discrete:
        ctl.add("base", [], "discrete.")  # nocoverage

    ctl.ground()
    with ctl.solve(yield_=True) as handle:
        facts = [str(s) + "." for s in handle.model().symbols(shown=True)]

    return facts


def check_user_input(facts: List[str]) -> None:
    """
    Checks if the user input is valid and returns a clingo.SolveResult
    """
    ctl = Control(message_limit=0)
    ctl.load(get_encoding("user-check.lp"))
    ctl.add("".join(facts))
    ctl.ground()
    with ctl.solve(yield_=True) as handle:
        warnings = [_parse_user_input_warnings(s) for s in handle.model().symbols(shown=True)]

    if warnings != []:
        msg = "Invalid user input.\n" + "\n".join(warnings)
        # raise ValueError(error_msg)
        # warn(msg)
        log.warning(msg)


def _parse_user_input_warnings(warning: Symbol) -> str:
    """
    Parses the warning/2 predicates of the user input check
    """
    warning_type = warning.arguments[0].string
    info = warning.arguments[1]

    match warning_type:
        case "not exists":
            variable = info.string
            return f'Variable "{variable}" does not exist.'
        # case "not part":
        #     variable = info.string
        #     return f"Variable {variable} cannot be added: Not a part."
        case "not attribute":
            variable = info.string
            return f'No value can be set for variable "{variable}". Variable exists but is a part.'
        case "outside domain":
            variable = info.arguments[0].string
            if str(info.arguments[1].type) == "SymbolType.Number":
                value = str(info.arguments[1].number)
            else:
                value = info.arguments[1].string
            return f'Value "{value}" is not in domain of variable "{variable}".'
        case _:  # nocoverage
            raise ValueError(f"Unknown warning type: {warning_type}")



"""
Test cases for solving.
"""

from typing import List, Optional
from unittest import TestCase

from . import run_test, unpack_test
from .clintests.tests_solve import TESTS_SOLVE


class TestClingo(TestCase):
    """
    Test cases for the clingo encoding.
    """

    def run_test(self, test_name: str, extra_ctl_args: Optional[List[str]] = None) -> None:
        """
        Runs a clintest test with the clingo encoding.
        """
        ctl_args = ["0"]
        if extra_ctl_args:
            ctl_args += extra_ctl_args
        test, program, files = unpack_test(test_name, TESTS_SOLVE)
        run_test(test, files=files, program=program, ctl_args=ctl_args, solver="clingo")

    def test_structure(self) -> None:
        """
        Test structure generation (clingo).
        """
        self.run_test("empty")
        self.run_test("optional_part")
        self.run_test("mandatory_part")
        self.run_test("part_with_cardinality")
        self.run_test("optional_part_with_subpart")

    def test_attributes(self) -> None:
        """
        Test attribute generation (clingo).
        """
        self.run_test("simple_discrete")
        self.run_test("optional_discrete")
        self.run_test("multiple_discrete")

        self.run_test("simple_integer")
        self.run_test("optional_integer")
        self.run_test("multiple_integer")

    def test_boolean_constraints(self) -> None:
        """
        Test Boolean constraints (clingo).
        """
        self.run_test("eq_sat")
        self.run_test("neq_sat")
        self.run_test("le_sat")
        self.run_test("leq_sat")
        self.run_test("ge_sat")
        self.run_test("geq_sat")

        self.run_test("eq_unsat")
        self.run_test("neq_unsat")
        self.run_test("le_unsat")
        self.run_test("leq_unsat")
        self.run_test("ge_unsat")
        self.run_test("geq_unsat")

        self.run_test("par_sat")
        self.run_test("neg_sat")
        self.run_test("or_sat")
        self.run_test("and_sat")

        self.run_test("par_unsat")
        self.run_test("neg_unsat")
        self.run_test("or_unsat")
        self.run_test("and_unsat")

        self.run_test("binary_undef")
        self.run_test("unary_undef")

    def test_table_constraints(self) -> None:
        """
        Test table constraints (clingo).
        """
        self.run_test("table_discrete")
        self.run_test("table_wildcard")
        self.run_test("table_integer")
        self.run_test("table_mixed")
        self.run_test("table_undef")
        self.run_test("table_undef2")
        self.run_test("empty_table")

    def test_arithmetics(self) -> None:
        """
        Test arithmetic formulas (clingo).
        """
        self.run_test("plus_sat")
        self.run_test("minus_sat")
        self.run_test("mult_sat")
        self.run_test("unary_plus_sat")
        self.run_test("unary_minus_sat")

        self.run_test("plus_unsat")
        self.run_test("minus_unsat")
        self.run_test("mult_unsat")
        self.run_test("unary_minus_unsat")

        self.run_test("plus_undef_sat")
        self.run_test("minus_undef_sat")
        self.run_test("plus_undef_unsat")
        self.run_test("minus_undef_unsat")

        self.run_test("precedence_sat")
        self.run_test("precedence_par_sat")
        self.run_test("precedence_unsat")
        self.run_test("precedence_par_unsat")

    def test_aggregates(self) -> None:
        """
        Test aggregation functions (clingo).
        """
        self.run_test("count")
        self.run_test("sum")
        self.run_test("min")
        self.run_test("max")

    def test_defaults(self) -> None:
        """
        Test solving defaults (clingo)
        """
        self.run_test("simple_default")
        self.run_test("simple_default_user")
        self.run_test("simple_default_include")

    def test_optimization(self) -> None:
        """
        Test solving optimization statements (clingo)
        """
        self.run_test("minimize", extra_ctl_args=["--opt-mode=optN"])
        self.run_test("maximize", extra_ctl_args=["--opt-mode=optN"])
        self.run_test("minimize_priority", extra_ctl_args=["--opt-mode=optN"])
        self.run_test("maximize_priority", extra_ctl_args=["--opt-mode=optN"])
        self.run_test("minimize_maximize_function", extra_ctl_args=["--opt-mode=optN"])
        self.run_test("maximize_minimize_function", extra_ctl_args=["--opt-mode=optN"])

    def test_user_input(self) -> None:
        """
        Test solving user input (clingo)
        """
        self.run_test("add_part")
        self.run_test("add_attribute")
        self.run_test("set_value_discrete")
        self.run_test("set_value_integer")

        self.run_test("set_invalid_variable")
        self.run_test("add_invalid_variable")
        self.run_test("set_invalid_type")
        self.run_test("set_invalid_value_discrete")
        self.run_test("set_invalid_value_num")


class TestFlingo(TestCase):
    """
    Test cases for the flingo encoding.
    """

    def run_test(self, test_name: str, extra_ctl_args: Optional[List[str]] = None) -> None:
        """
        Runs a clintest test with the flingo encoding.
        """
        ctl_args = ["0"]
        if extra_ctl_args:
            ctl_args += extra_ctl_args  # nocoverage
        test, program, files = unpack_test(test_name, TESTS_SOLVE, flingo=True)
        run_test(test, files=files, program=program, ctl_args=ctl_args, solver="flingo", preprocess="False")

    def test_structure(self) -> None:
        """
        Test structure generation (flingo).
        """
        self.run_test("empty")
        self.run_test("optional_part")
        self.run_test("mandatory_part")
        self.run_test("part_with_cardinality")
        self.run_test("optional_part_with_subpart")

    def test_attributes(self) -> None:
        """
        Test attribute generation (flingo).
        """
        self.run_test("simple_discrete")
        self.run_test("optional_discrete")
        self.run_test("multiple_discrete")

        self.run_test("simple_integer")
        self.run_test("optional_integer")
        self.run_test("multiple_integer")
        self.run_test("unbounded_integer")
        self.run_test("unbounded_integer_below")
        self.run_test("unbounded_integer_above")

    def test_boolean_constraints(self) -> None:
        """
        Test Boolean constraints (flingo).
        """
        self.run_test("eq_sat")
        self.run_test("neq_sat")
        self.run_test("le_sat")
        self.run_test("leq_sat")
        self.run_test("ge_sat")
        self.run_test("geq_sat")

        self.run_test("eq_unsat")
        self.run_test("neq_unsat")
        self.run_test("le_unsat")
        self.run_test("leq_unsat")
        self.run_test("ge_unsat")
        self.run_test("geq_unsat")

        self.run_test("par_sat")
        self.run_test("neg_sat")
        self.run_test("or_sat")
        self.run_test("and_sat")

        self.run_test("par_unsat")
        self.run_test("neg_unsat")
        self.run_test("or_unsat")
        self.run_test("and_unsat")

        self.run_test("binary_undef")
        self.run_test("unary_undef")

    def test_table_constraints(self) -> None:
        """
        Test table constraints (flingo).
        """
        self.run_test("table_discrete")
        self.run_test("table_wildcard")
        self.run_test("table_integer")
        self.run_test("table_mixed")
        self.run_test("table_undef")
        self.run_test("table_undef2")
        self.run_test("empty_table")

    def test_arithmetics(self) -> None:
        """
        Test arithmetic formulas (flingo).
        """
        self.run_test("plus_sat")
        self.run_test("minus_sat")
        self.run_test("unary_plus_sat")
        self.run_test("unary_minus_sat")

        self.run_test("plus_unsat")
        self.run_test("minus_unsat")
        self.run_test("unary_minus_unsat")

        self.run_test("plus_undef_sat")
        self.run_test("minus_undef_sat")
        self.run_test("plus_undef_unsat")
        self.run_test("minus_undef_unsat")

        self.run_test("precedence_sat")
        self.run_test("precedence_par_sat")
        self.run_test("precedence_unsat")
        self.run_test("precedence_par_unsat")

    def test_aggregates(self) -> None:
        """
        Test aggregation functions (flingo).
        """
        self.run_test("count")
        self.run_test("sum")
        self.run_test("min")
        self.run_test("max")

    def test_defaults(self) -> None:
        """
        Test solving defaults (flingo)
        """
        self.run_test("simple_default")
        self.run_test("simple_default_user")
        self.run_test("simple_default_include")

    def test_optimization(self) -> None:
        """
        Test solving optimization statements (flingo)
        """
        self.run_test("minimize")
        self.run_test("maximize")
        # self.run_test("minimize_priority")
        # self.run_test("maximize_priority")
        # self.run_test("minimize_maximize_function")
        # self.run_test("maximize_minimize_function")

    def test_user_input(self) -> None:
        """
        Test solving user input (flingo)
        """
        self.run_test("add_part")
        self.run_test("add_attribute")
        self.run_test("set_value_discrete")
        self.run_test("set_value_integer")

        self.run_test("set_invalid_variable")
        self.run_test("add_invalid_variable")
        self.run_test("set_invalid_type")
        self.run_test("set_invalid_value_discrete")
        self.run_test("set_invalid_value_num")



"""
Setup project wide loggers.

This is a thin wrapper around Python's logging module. It supports colored
logging.
"""

import logging
from typing import TextIO

NOTSET = logging.NOTSET
DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

COLORS = {
    "GREY": "\033[90m",
    "BLUE": "\033[94m",
    "GREEN": "\033[92m",
    "YELLOW": "\033[93m",
    "RED": "\033[91m",
    "NORMAL": "\033[0m",
}


class SingleLevelFilter(logging.Filter):
    """
    Filter levels.
    """

    passlevel: int
    reject: bool

    def __init__(self, passlevel: int, reject: bool):
        # pylint: disable=super-init-not-called
        self.passlevel = passlevel
        self.reject = reject

    def filter(self, record: logging.LogRecord) -> bool:
        if self.reject:
            return record.levelno != self.passlevel  # nocoverage

        return record.levelno == self.passlevel


def configure_logging(stream: TextIO, level: int, use_color: bool) -> None:
    """
    Configure application logging.
    """

    def format_str(color: str) -> str:
        if use_color:
            return f"{COLORS[color]}%(levelname)s:{COLORS['GREY']}  - %(message)s{COLORS['NORMAL']}"
        return "%(levelname)s:  - %(message)s"  # nocoverage

    def make_handler(level: int, color: str) -> "logging.StreamHandler[TextIO]":
        handler = logging.StreamHandler(stream)
        handler.addFilter(SingleLevelFilter(level, False))
        handler.setLevel(level)
        formatter = logging.Formatter(format_str(color))
        handler.setFormatter(formatter)
        return handler

    handlers = [
        make_handler(logging.INFO, "GREEN"),
        make_handler(logging.WARNING, "YELLOW"),
        make_handler(logging.DEBUG, "BLUE"),
        make_handler(logging.ERROR, "RED"),
    ]
    logging.basicConfig(handlers=handlers, level=level)


def get_logger(name: str) -> logging.Logger:
    """
    Get a logger with the given name.
    """
    return logging.getLogger(name)



"""
The coomsuite project.
"""

from tempfile import NamedTemporaryFile
from typing import List, Optional, Tuple

from antlr4 import FileStream
from clingo.application import clingo_main

from .application import COOMSolverApp
from .preprocess import check_user_input, preprocess
from .utils import run_antlr4_visitor
from .utils.logging import get_logger

log = get_logger("main")

SOLVERS = ["clingo", "flingo"]


def write_facts(facts: str, outfile: str) -> None:  # nocoverage
    """
    Auxiilary function to write ASP facts to a file

    Args:
        facts (str): ASP facts
        outfile (str): Output file
    """
    with open(outfile, "w", encoding="utf8") as f:
        f.write("%%% serialized COOM facts \n")
        f.write(facts)
    log.info("ASP file saved in %s", outfile)


def convert_coom(coom_model: str, coom_user: Optional[str] = None) -> Tuple[str, bool, bool]:  # nocoverage
    """
    Converts a COOM instance into ASP
    Args:
        coom_model (str): COOM model file .coom
        coom_user (str, optional): COOM user input file .coom
    """
    unbounded = False
    optimize = False

    input_stream_model = FileStream(coom_model, encoding="utf-8")
    asp_instance = "\n".join(
        [f"coom_{a}" if a != "" else a for a in run_antlr4_visitor(input_stream_model, grammar="model")]
    )

    if "#sup" in asp_instance:
        unbounded = True
    if "coom_minimize" in asp_instance or "coom_maximize" in asp_instance:
        optimize = True
    if coom_user is not None:
        input_stream_user = FileStream(coom_user, encoding="utf-8")
        asp_instance += "\n".join(
            [f"coom_{a}" if a != "" else a for a in run_antlr4_visitor(input_stream_user, grammar="user")]
        )

    return asp_instance, unbounded, optimize


def solve(
    serialized_facts: List[str],
    solver: str,
    max_bound: int,
    clingo_args: List[str],
    output: str,
) -> int:  # nocoverage
    """
    Preprocesses and solves a serialized COOM instance.
    """
    # Preprocess serialized ASP facts
    processed_facts = preprocess(
        serialized_facts,
        max_bound=max_bound,
        discrete=solver == "clingo",
    )
    check_user_input(processed_facts)

    with NamedTemporaryFile(mode="w", delete=False) as tmp:
        tmp_name = tmp.name
        tmp.write("".join(processed_facts))

    # Solve the ASP instance
    return clingo_main(
        COOMSolverApp(
            options={
                "solver": solver,
                "output_format": output,
            }
        ),
        [tmp_name] + clingo_args,
    )



# Generated from /home/heuer/tools/coom-suite/src/coomsuite/utils/coom_grammar/model/Model.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .ModelParser import ModelParser
else:
    from ModelParser import ModelParser

# This class defines a complete generic visitor for a parse tree produced by ModelParser.

class ModelVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by ModelParser#root.
    def visitRoot(self, ctx:ModelParser.RootContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#product.
    def visitProduct(self, ctx:ModelParser.ProductContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#structure.
    def visitStructure(self, ctx:ModelParser.StructureContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#enumeration.
    def visitEnumeration(self, ctx:ModelParser.EnumerationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#feature.
    def visitFeature(self, ctx:ModelParser.FeatureContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#cardinality.
    def visitCardinality(self, ctx:ModelParser.CardinalityContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#attribute.
    def visitAttribute(self, ctx:ModelParser.AttributeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#option.
    def visitOption(self, ctx:ModelParser.OptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#field.
    def visitField(self, ctx:ModelParser.FieldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#number_def.
    def visitNumber_def(self, ctx:ModelParser.Number_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#string_def.
    def visitString_def(self, ctx:ModelParser.String_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#unit.
    def visitUnit(self, ctx:ModelParser.UnitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#fraction.
    def visitFraction(self, ctx:ModelParser.FractionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#behavior.
    def visitBehavior(self, ctx:ModelParser.BehaviorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#behavior_block.
    def visitBehavior_block(self, ctx:ModelParser.Behavior_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#define.
    def visitDefine(self, ctx:ModelParser.DefineContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#conditioned.
    def visitConditioned(self, ctx:ModelParser.ConditionedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#exists.
    def visitExists(self, ctx:ModelParser.ExistsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#explanation.
    def visitExplanation(self, ctx:ModelParser.ExplanationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#precondition.
    def visitPrecondition(self, ctx:ModelParser.PreconditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#date.
    def visitDate(self, ctx:ModelParser.DateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#combinations.
    def visitCombinations(self, ctx:ModelParser.CombinationsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#combination_row.
    def visitCombination_row(self, ctx:ModelParser.Combination_rowContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#combination_item.
    def visitCombination_item(self, ctx:ModelParser.Combination_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#combination_atom.
    def visitCombination_atom(self, ctx:ModelParser.Combination_atomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#assign_default.
    def visitAssign_default(self, ctx:ModelParser.Assign_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#assign_imply.
    def visitAssign_imply(self, ctx:ModelParser.Assign_implyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#optimize.
    def visitOptimize(self, ctx:ModelParser.OptimizeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#interaction.
    def visitInteraction(self, ctx:ModelParser.InteractionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#message.
    def visitMessage(self, ctx:ModelParser.MessageContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#require.
    def visitRequire(self, ctx:ModelParser.RequireContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#prefer.
    def visitPrefer(self, ctx:ModelParser.PreferContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#condition.
    def visitCondition(self, ctx:ModelParser.ConditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#condition_or.
    def visitCondition_or(self, ctx:ModelParser.Condition_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#condition_and.
    def visitCondition_and(self, ctx:ModelParser.Condition_andContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#condition_not.
    def visitCondition_not(self, ctx:ModelParser.Condition_notContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#condition_compare.
    def visitCondition_compare(self, ctx:ModelParser.Condition_compareContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#condition_part.
    def visitCondition_part(self, ctx:ModelParser.Condition_partContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#formula.
    def visitFormula(self, ctx:ModelParser.FormulaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#formula_add.
    def visitFormula_add(self, ctx:ModelParser.Formula_addContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#formula_sub.
    def visitFormula_sub(self, ctx:ModelParser.Formula_subContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#formula_mul.
    def visitFormula_mul(self, ctx:ModelParser.Formula_mulContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#formula_div.
    def visitFormula_div(self, ctx:ModelParser.Formula_divContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#formula_pow.
    def visitFormula_pow(self, ctx:ModelParser.Formula_powContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#formula_sign.
    def visitFormula_sign(self, ctx:ModelParser.Formula_signContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#formula_func.
    def visitFormula_func(self, ctx:ModelParser.Formula_funcContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#formula_atom.
    def visitFormula_atom(self, ctx:ModelParser.Formula_atomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#constant.
    def visitConstant(self, ctx:ModelParser.ConstantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#floating.
    def visitFloating(self, ctx:ModelParser.FloatingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#path.
    def visitPath(self, ctx:ModelParser.PathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#path_item.
    def visitPath_item(self, ctx:ModelParser.Path_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#path_index.
    def visitPath_index(self, ctx:ModelParser.Path_indexContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#name.
    def visitName(self, ctx:ModelParser.NameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#stmt_end.
    def visitStmt_end(self, ctx:ModelParser.Stmt_endContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ModelParser#compare.
    def visitCompare(self, ctx:ModelParser.CompareContext):
        return self.visitChildren(ctx)



del ModelParser



"""
The command line parser for the project.
"""

import sys
from argparse import ArgumentParser
from textwrap import dedent
from typing import Any, Optional, cast

from coomsuite import SOLVERS

from . import logging

__all__ = ["get_parser"]

if sys.version_info[1] < 8:
    import importlib_metadata as metadata  # nocoverage
else:
    from importlib import metadata  # nocoverage

VERSION = metadata.version("coomsuite")


def get_parser() -> ArgumentParser:
    """
    Return the parser for command line options.
    """
    parser = ArgumentParser(
        prog="coomsuite",
        description=dedent(
            """\
            The COOM suite is a package providing functionality
            to parse and solve product configuration problems specified in COOM.
            """
        ),
    )
    levels = [
        ("error", logging.ERROR),
        ("warning", logging.WARNING),
        ("info", logging.INFO),
        ("debug", logging.DEBUG),
    ]

    def get(levels: list[tuple[str, int]], name: str) -> Optional[int]:
        for key, val in levels:
            if key == name:
                return val
        return None  # nocoverage

    parser.add_argument(
        "--log",
        default="warning",
        choices=[val for _, val in levels],
        metavar=f'{{{",".join(key for key, _ in levels)}}}',
        help="set log level [%(default)s]",
        type=cast(Any, lambda name: get(levels, name)),
    )

    parser.add_argument("--version", "-v", action="version", version=f"%(prog)s {VERSION}")

    subparsers = parser.add_subparsers(help="Sub commands", dest="command")

    # -------------
    #  Instance parser
    # -------------
    parser_convert = subparsers.add_parser("convert", help="Convert COOM instance to ASP.")
    parser_convert.add_argument(
        "input",
        type=str,
        help="""
        Input the COOM model file corresponding to the instance. Converted instance is printed to console by default.
        Specify output directory with '--output' to save .""",
    )
    parser_convert.add_argument(
        "--output",
        "-o",
        type=str,
        default=None,
        help="Path to output directory. (Optional)",
    )

    parser_convert.add_argument("--user-input", "-u", type=str, help="Input the COOM user input.")
    # -------------
    # Solve parser
    # -------------
    parser_solve = subparsers.add_parser(
        "solve",
        help="Converts and solves the COOM instance.",
        epilog="In addition to the above commands, all usual clingo commands can be used, e.g., -n0 to print all models.",  # pylint: disable=line-too-long
    )
    parser_solve.add_argument(
        "input",
        type=str,
        help="Path to the COOM model file to solve",
    )
    parser_solve.add_argument("--user-input", "-u", type=str, help="Input the COOM user input.")

    parser_solve.add_argument("--solver", "-s", type=str, help="Set solver", choices=SOLVERS, default="clingo")

    parser_solve.add_argument(
        "--output", "-o", type=str, help="Set console output format", choices=["asp", "coom"], default="asp"
    )
    parser_solve.add_argument("--show-facts", action="store_true", help="Show preprocessed fact format")
    parser_solve.add_argument(
        "--bounds",
        "-b",
        type=str,
        help="Solve with search for bounds for problems with unbounded cardinalities.",
        choices=["linear", "exponential"],
        default="linear",
    )
    parser_solve.add_argument(
        "--multishot",
        "-m",
        action="store_true",
        help="""
        Use multi-shot solving for problems with unbounded cardinalities.
        (If --bounds is not used this flag is ignored.)""",
    )
    parser_solve.add_argument(
        "--initial-bound",
        "-i",
        type=int,
        default=0,
        help="Initial bound for the incremental search algorithm. Default is 0.",
    )
    parser_solve.add_argument(
        "--step",
        type=int,
        help="""
        Parameter for getting the next bound in incremental search algorithm:
            - For --bounds=linear, this is the linear increment. Default is 1.
            - For --bounds=exponential, this is the base of the exponential. Default is 2.""",
    )
    return parser



"""
Contains a dictionary with all clintest tests for preprocessing
and the corresponding files (usually corresponds to "testname.lp")or programs they should be run with.

The key of the dictionary corresponds to the name of the test.

All tests run with clingo.
"""

# pylint: disable=line-too-long, too-many-lines
from typing import Any

from .. import get_model_from_file
from . import TEST_EMPTY, StableModels, Supersets

TESTS_PREPROCESS: dict[str, dict[str, Any]] = {
    "empty": {"test": TEST_EMPTY, "program": ""},
    "empty_product": {
        "test": StableModels({'part("product")', 'type("root","product")'}),
        "program": 'coom_structure("product").',
    },
    "structure_mandatory": {
        "test": StableModels(
            {
                'part("product")',
                'part("Wheel")',
                'constraint(("root.wheel",1),"lowerbound")',
                'index("root.wheel[0]",0)',
                'parent("root.wheel[0]","root")',
                'set("root.wheel","root.wheel[0]")',
                'type("root","product")',
                'type("root.wheel[0]","Wheel")',
            }
        ),
        "program": """
            coom_structure("product").
            coom_feature("product","wheel","Wheel",1,1).
            coom_structure("Wheel").""",
    },
    "structure_optional": {
        "test": StableModels(
            {
                'part("product")',
                'part("Wheel")',
                'constraint(("root.wheel",0),"lowerbound")',
                'index("root.wheel[0]",0)',
                'parent("root.wheel[0]","root")',
                'set("root.wheel","root.wheel[0]")',
                'type("root","product")',
                'type("root.wheel[0]","Wheel")',
            }
        ),
        "program": """
            coom_structure("product").
            coom_feature("product","wheel","Wheel",0,1).
            coom_structure("Wheel").""",
    },
    "structure_nested": {
        "test": StableModels(
            {
                'part("product")',
                'part("Carrier")',
                'part("Bag")',
                'constraint(("root.carrier",1),"lowerbound")',
                'constraint(("root.carrier[0].bag",1),"lowerbound")',
                'index("root.carrier[0]",0)',
                'index("root.carrier[0].bag[0]",0)',
                'parent("root.carrier[0]","root")',
                'parent("root.carrier[0].bag[0]","root.carrier[0]")',
                'set("root.carrier","root.carrier[0]")',
                'set("root.carrier[0].bag","root.carrier[0].bag[0]")',
                'type("root","product")',
                'type("root.carrier[0]","Carrier")',
                'type("root.carrier[0].bag[0]","Bag")',
            }
        ),
        "program": """
            coom_structure("product").
            coom_feature("product","carrier","Carrier",1,1).
            coom_structure("Carrier").
            coom_feature("Carrier","bag","Bag",1,1).
            coom_structure("Bag").""",
    },
    "structure_nested_optional": {
        "test": StableModels(
            {
                'part("product")',
                'part("Carrier")',
                'part("Bag")',
                'constraint(("root.carrier",0),"lowerbound")',
                'constraint(("root.carrier[0].bag",0),"lowerbound")',
                'index("root.carrier[0]",0)',
                'index("root.carrier[0].bag[0]",0)',
                'index("root.carrier[0].bag[1]",1)',
                'parent("root.carrier[0]","root")',
                'parent("root.carrier[0].bag[0]","root.carrier[0]")',
                'parent("root.carrier[0].bag[1]","root.carrier[0]")',
                'set("root.carrier","root.carrier[0]")',
                'set("root.carrier[0].bag","root.carrier[0].bag[0]")',
                'set("root.carrier[0].bag","root.carrier[0].bag[1]")',
                'type("root","product")',
                'type("root.carrier[0]","Carrier")',
                'type("root.carrier[0].bag[0]","Bag")',
                'type("root.carrier[0].bag[1]","Bag")',
            }
        ),
        "program": """
            coom_structure("product").
            coom_feature("product","carrier","Carrier",0,1).
            coom_structure("Carrier").
            coom_feature("Carrier","bag","Bag",0,2).
            coom_structure("Bag").""",
    },
    "enumeration": {
        "test": StableModels(
            {
                'discrete("Color")',
                'part("product")',
                'constraint(("root.color",1),"lowerbound")',
                'domain("Color","Red")',
                'domain("Color","Green")',
                'domain("Color","Blue")',
                'index("root.color[0]",0)',
                'parent("root.color[0]","root")',
                'set("root.color","root.color[0]")',
                'type("root","product")',
                'type("root.color[0]","Color")',
            }
        ),
        "program": """
            coom_structure("product").
            coom_feature("product","color","Color",1,1).
            coom_enumeration("Color").
            coom_option("Color", "Red").
            coom_option("Color", "Green").
            coom_option("Color", "Blue").""",
    },
    "bool_enumeration": {
        "test": StableModels(
            {
                'discrete("Bool")',
                'part("product")',
                'constraint(("root.boolean",1),"lowerbound")',
                'domain("Bool","True")',
                'domain("Bool","False")',
                'index("root.boolean[0]",0)',
                'parent("root.boolean[0]","root")',
                'set("root.boolean","root.boolean[0]")',
                'type("root","product")',
                'type("root.boolean[0]","Bool")',
            }
        ),
        "program": """
            coom_structure("product").
            coom_feature("product","boolean","Bool",1,1).""",
    },
    "attribute": {
        "test": StableModels(
            {
                'discrete("Wheel")',
                'integer("Wheel.size")',
                'part("product")',
                'constraint(("root.wheel",1),"lowerbound")',
                'constraint(("root.wheel[0].size",1),"lowerbound")',
                'constraint(("Wheel","root.wheel[0]"),"table")',
                'domain("Wheel","W14")',
                'index("root.wheel[0]",0)',
                'index("root.wheel[0].size[0]",0)',
                'parent("root.wheel[0]","root")',
                'parent("root.wheel[0].size[0]","root.wheel[0]")',
                'set("root.wheel","root.wheel[0]")',
                'set("root.wheel[0].size","root.wheel[0].size[0]")',
                'type("root","product")',
                'type("root.wheel[0]","Wheel")',
                'type("root.wheel[0].size[0]","Wheel.size")',
                'allow("Wheel",(0,0),"W14")',
                'allow("Wheel",(1,0),14)',
                'range("Wheel.size",14,14)',
                'column(("Wheel","root.wheel[0]"),0,1,"root.wheel[0].size[0]")',
                'column(("Wheel","root.wheel[0]"),0,0,"root.wheel[0]")',
            }
        ),
        "program": """
            coom_structure("product").
            coom_feature("product","wheel","Wheel",1,1).
            coom_enumeration("Wheel").
            coom_attribute("Wheel","size","num").
            coom_option("Wheel", "W14").
            coom_attribute_value("Wheel","W14","size",14).""",
    },
    "require_undef": {
        "test": StableModels({'constant("Silver")', 'part("product")', 'type("root","product")'}),
        "program": """
            coom_structure("product").

            coom_behavior(("product",0)).
            coom_require(("product",0),"color=Silver").
            coom_binary("color=Silver","color","=","Silver").
            coom_path("color",0,"color").
            coom_constant("Silver").""",
    },
    "require_with_number": {
        "test": StableModels(
            {
                'discrete("Wheel")',
                'integer("Wheel.size")',
                'part("product")',
                'constraint(("root.wheel",1),"lowerbound")',
                'constraint(("root.wheel[0].size",1),"lowerbound")',
                'constraint(("Wheel","root.wheel[0]"),"table")',
                'constraint((0,"root.wheel[0].size[0]=27"),"boolean")',
                'domain("Wheel","W27")',
                'domain("Wheel","W28")',
                'index("root.wheel[0]",0)',
                'index("root.wheel[0].size[0]",0)',
                'number("27",27)',
                'parent("root.wheel[0]","root")',
                'parent("root.wheel[0].size[0]","root.wheel[0]")',
                'set("root.wheel","root.wheel[0]")',
                'set("root.wheel[0].size","root.wheel[0].size[0]")',
                'type("root","product")',
                'type("root.wheel[0]","Wheel")',
                'type("root.wheel[0].size[0]","Wheel.size")',
                'allow("Wheel",(0,0),"W27")',
                'allow("Wheel",(0,1),"W28")',
                'allow("Wheel",(1,1),28)',
                'allow("Wheel",(1,0),27)',
                'range("Wheel.size",27,28)',
                'binary("root.wheel[0].size[0]=27","root.wheel[0].size[0]","=","27")',
                'column(("Wheel","root.wheel[0]"),0,1,"root.wheel[0].size[0]")',
                'column(("Wheel","root.wheel[0]"),0,0,"root.wheel[0]")',
            }
        ),
    },
    "require_with_number_ge": {
        "test": StableModels(
            {
                'discrete("Wheel")',
                'integer("Wheel.size")',
                'part("product")',
                'constraint(("root.wheel",1),"lowerbound")',
                'constraint(("root.wheel[0].size",1),"lowerbound")',
                'constraint(("Wheel","root.wheel[0]"),"table")',
                'constraint((0,"root.wheel[0].size[0]>=28"),"boolean")',
                'domain("Wheel","W27")',
                'domain("Wheel","W28")',
                'index("root.wheel[0]",0)',
                'index("root.wheel[0].size[0]",0)',
                'number("28",28)',
                'parent("root.wheel[0]","root")',
                'parent("root.wheel[0].size[0]","root.wheel[0]")',
                'set("root.wheel","root.wheel[0]")',
                'set("root.wheel[0].size","root.wheel[0].size[0]")',
                'type("root","product")',
                'type("root.wheel[0]","Wheel")',
                'type("root.wheel[0].size[0]","Wheel.size")',
                'allow("Wheel",(0,0),"W27")',
                'allow("Wheel",(0,1),"W28")',
                'allow("Wheel",(1,1),28)',
                'allow("Wheel",(1,0),27)',
                'range("Wheel.size",27,28)',
                'binary("root.wheel[0].size[0]>=28","root.wheel[0].size[0]",">=","28")',
                'column(("Wheel","root.wheel[0]"),0,1,"root.wheel[0].size[0]")',
                'column(("Wheel","root.wheel[0]"),0,0,"root.wheel[0]")',
            }
        ),
    },
    "require_with_constant": {
        "test": StableModels(
            {
                'constant("W28")',
                'discrete("Wheel")',
                'part("product")',
                'constraint(("root.wheel",1),"lowerbound")',
                'constraint((0,"root.wheel[0]=W28"),"boolean")',
                'domain("Wheel","W27")',
                'domain("Wheel","W28")',
                'index("root.wheel[0]",0)',
                'parent("root.wheel[0]","root")',
                'set("root.wheel","root.wheel[0]")',
                'type("root","product")',
                'type("root.wheel[0]","Wheel")',
                'binary("root.wheel[0]=W28","root.wheel[0]","=","W28")',
            }
        ),
    },
    "require_two_wheels": {
        "test": StableModels(
            {
                'discrete("Wheel")',
                'integer("Wheel.size")',
                'part("product")',
                'constraint(("root.frontWheel",1),"lowerbound")',
                'constraint(("root.rearWheel",1),"lowerbound")',
                'constraint(("root.frontWheel[0].size",1),"lowerbound")',
                'constraint(("root.rearWheel[0].size",1),"lowerbound")',
                'constraint(("Wheel","root.frontWheel[0]"),"table")',
                'constraint(("Wheel","root.rearWheel[0]"),"table")',
                'constraint((0,"root.frontWheel[0].size[0]=root.rearWheel[0].size[0]"),"boolean")',
                'domain("Wheel","W27")',
                'domain("Wheel","W28")',
                'index("root.frontWheel[0]",0)',
                'index("root.rearWheel[0]",0)',
                'index("root.rearWheel[0].size[0]",0)',
                'index("root.frontWheel[0].size[0]",0)',
                'parent("root.frontWheel[0]","root")',
                'parent("root.rearWheel[0]","root")',
                'parent("root.rearWheel[0].size[0]","root.rearWheel[0]")',
                'parent("root.frontWheel[0].size[0]","root.frontWheel[0]")',
                'set("root.frontWheel","root.frontWheel[0]")',
                'set("root.rearWheel","root.rearWheel[0]")',
                'set("root.frontWheel[0].size","root.frontWheel[0].size[0]")',
                'set("root.rearWheel[0].size","root.rearWheel[0].size[0]")',
                'type("root","product")',
                'type("root.frontWheel[0]","Wheel")',
                'type("root.rearWheel[0]","Wheel")',
                'type("root.frontWheel[0].size[0]","Wheel.size")',
                'type("root.rearWheel[0].size[0]","Wheel.size")',
                'allow("Wheel",(0,0),"W27")',
                'allow("Wheel",(0,1),"W28")',
                'allow("Wheel",(1,1),28)',
                'allow("Wheel",(1,0),27)',
                'range("Wheel.size",27,28)',
                'binary("root.frontWheel[0].size[0]=root.rearWheel[0].size[0]","root.frontWheel[0].size[0]","=","root.rearWheel[0].size[0]")',
                'column(("Wheel","root.frontWheel[0]"),0,1,"root.frontWheel[0].size[0]")',
                'column(("Wheel","root.rearWheel[0]"),0,1,"root.rearWheel[0].size[0]")',
                'column(("Wheel","root.frontWheel[0]"),0,0,"root.frontWheel[0]")',
                'column(("Wheel","root.rearWheel[0]"),0,0,"root.rearWheel[0]")',
            }
        ),
    },
    "conditional_require": {
        "test": StableModels(
            {
                'constant("True")',
                'constant("Small")',
                'discrete("Wheel")',
                'discrete("Bool")',
                'part("product")',
                'constraint(("root.wheel",1),"lowerbound")',
                'constraint(("root.wheelSupport",1),"lowerbound")',
                'constraint((0,"!root.wheelSupport[0]=True||root.wheel[0]=Small"),"boolean")',
                'domain("Wheel","Small")',
                'domain("Wheel","Large")',
                'domain("Bool","True")',
                'domain("Bool","False")',
                'index("root.wheel[0]",0)',
                'index("root.wheelSupport[0]",0)',
                'parent("root.wheel[0]","root")',
                'parent("root.wheelSupport[0]","root")',
                'set("root.wheel","root.wheel[0]")',
                'set("root.wheelSupport","root.wheelSupport[0]")',
                'type("root","product")',
                'type("root.wheel[0]","Wheel")',
                'type("root.wheelSupport[0]","Bool")',
                'unary("!root.wheelSupport[0]=True","!","root.wheelSupport[0]=True")',
                'binary("root.wheel[0]=Small","root.wheel[0]","=","Small")',
                'binary("root.wheelSupport[0]=True","root.wheelSupport[0]","=","True")',
                'binary("!root.wheelSupport[0]=True||root.wheel[0]=Small","!root.wheelSupport[0]=True","||","root.wheel[0]=Small")',
            }
        ),
    },
    "multiple_conditions": {
        "test": StableModels(
            {
                'discrete("Color")',
                'discrete("Size")',
                'discrete("Material")',
                'domain("Color","Red")',
                'domain("Color","Green")',
                'domain("Size","Small")',
                'domain("Size","Medium")',
                'domain("Material","Cotton")',
                'domain("Material","Polyester")',
                'type("root","product")',
                'type("root.color[0]","Color")',
                'type("root.size[0]","Size")',
                'type("root.material[0]","Material")',
                'index("root.color[0]",0)',
                'index("root.size[0]",0)',
                'index("root.material[0]",0)',
                'parent("root.color[0]","root")',
                'parent("root.size[0]","root")',
                'parent("root.material[0]","root")',
                'constraint(("root.color",1),"lowerbound")',
                'constraint(("root.size",1),"lowerbound")',
                'constraint(("root.material",1),"lowerbound")',
                'constraint((0,"!(root.size[0]=Small&&root.material[0]=Cotton)||root.color[0]=Red"),"boolean")',
                'binary("root.color[0]=Red","root.color[0]","=","Red")',
                'binary("root.material[0]=Cotton","root.material[0]","=","Cotton")',
                'binary("root.size[0]=Small","root.size[0]","=","Small")',
                'binary("root.size[0]=Small&&root.material[0]=Cotton","root.size[0]=Small","&&","root.material[0]=Cotton")',
                'binary("!(root.size[0]=Small&&root.material[0]=Cotton)||root.color[0]=Red","!(root.size[0]=Small&&root.material[0]=Cotton)","||","root.color[0]=Red")',
                'unary("(root.size[0]=Small&&root.material[0]=Cotton)","()","root.size[0]=Small&&root.material[0]=Cotton")',
                'unary("!(root.size[0]=Small&&root.material[0]=Cotton)","!","(root.size[0]=Small&&root.material[0]=Cotton)")',
                'set("root.color","root.color[0]")',
                'set("root.size","root.size[0]")',
                'set("root.material","root.material[0]")',
                'part("product")',
                'constant("Small")',
                'constant("Cotton")',
                'constant("Red")',
            }
        ),
    },
    "require_with_optional_part": {
        "test": StableModels(
            {
                'integer("A.b")',
                'range("A.b",1,2)',
                'type("root","product")',
                'type("root.a[0]","A")',
                'type("root.a[0].b[0]","A.b")',
                'index("root.a[0]",0)',
                'index("root.a[0].b[0]",0)',
                'parent("root.a[0]","root")',
                'parent("root.a[0].b[0]","root.a[0]")',
                'constraint(("root.a",0),"lowerbound")',
                'constraint(("root.a[0].b",1),"lowerbound")',
                'constraint((0,"!count(root.a[0])=1||sum(root.a[0].b)>1"),"boolean")',
                'function("sum(root.a[0].b)","sum","root.a[0].b")',
                'function("count(root.a[0])","count","root.a[0]")',
                'set("root.a","root.a[0]")',
                'set("root.a[0].b","root.a[0].b[0]")',
                'set("root.a[0]","root.a[0]")',
                'binary("sum(root.a[0].b)>1","sum(root.a[0].b)",">","1")',
                'binary("count(root.a[0])=1","count(root.a[0])","=","1")',
                'binary("!count(root.a[0])=1||sum(root.a[0].b)>1","!count(root.a[0])=1","||","sum(root.a[0].b)>1")',
                'number("1",1)',
                'unary("!count(root.a[0])=1","!","count(root.a[0])=1")',
                'part("product")',
                'part("A")',
            }
        ),
    },
    "conditional_require_undef": {
        "test": StableModels(
            {
                'constant("Silver")',
                'constant("Big")',
                'discrete("Size")',
                'part("product")',
                'constraint(("root.Size",1),"lowerbound")',
                'domain("Size","Small")',
                'domain("Size","Big")',
                'index("root.Size[0]",0)',
                'parent("root.Size[0]","root")',
                'set("root.Size","root.Size[0]")',
                'type("root","product")',
                'type("root.Size[0]","size")',
            }
        ),
    },
    "require_multiple_instances": {
        "test": StableModels(
            {
                'constant("W28")',
                'discrete("Size")',
                'part("product")',
                'part("Wheel")',
                'constraint(("root.wheel",2),"lowerbound")',
                'constraint(("root.wheel[0].size",1),"lowerbound")',
                'constraint(("root.wheel[1].size",1),"lowerbound")',
                'constraint((0,"root.wheel[0].size[0]=W28"),"boolean")',
                'constraint((0,"root.wheel[1].size[0]=W28"),"boolean")',
                'domain("Size","W27")',
                'domain("Size","W28")',
                'index("root.wheel[0]",0)',
                'index("root.wheel[1]",1)',
                'index("root.wheel[1].size[0]",0)',
                'index("root.wheel[0].size[0]",0)',
                'parent("root.wheel[0]","root")',
                'parent("root.wheel[1]","root")',
                'parent("root.wheel[1].size[0]","root.wheel[1]")',
                'parent("root.wheel[0].size[0]","root.wheel[0]")',
                'set("root.wheel","root.wheel[0]")',
                'set("root.wheel","root.wheel[1]")',
                'set("root.wheel[0].size","root.wheel[0].size[0]")',
                'set("root.wheel[1].size","root.wheel[1].size[0]")',
                'type("root","product")',
                'type("root.wheel[0]","Wheel")',
                'type("root.wheel[1]","Wheel")',
                'type("root.wheel[1].size[0]","Size")',
                'type("root.wheel[0].size[0]","Size")',
                'binary("root.wheel[1].size[0]=W28","root.wheel[1].size[0]","=","W28")',
                'binary("root.wheel[0].size[0]=W28","root.wheel[0].size[0]","=","W28")',
            }
        ),
    },
    "require_with_partonomy": {
        "test": StableModels(
            {
                'constant("Red")',
                'discrete("Color")',
                'part("product")',
                'part("Basket")',
                'constraint(("root.basket",1),"lowerbound")',
                'constraint(("root.basket[0].color",1),"lowerbound")',
                'constraint((0,"!count(root.basket[0])=1||root.basket[0].color[0]=Red"),"boolean")',
                'domain("Color","Red")',
                'domain("Color","Yellow")',
                'index("root.basket[0]",0)',
                'index("root.basket[0].color[0]",0)',
                'parent("root.basket[0]","root")',
                'parent("root.basket[0].color[0]","root.basket[0]")',
                'set("root.basket","root.basket[0]")',
                'set("root.basket[0].color","root.basket[0].color[0]")',
                'type("root","product")',
                'type("root.basket[0]","Basket")',
                'type("root.basket[0].color[0]","Color")',
                'binary("root.basket[0].color[0]=Red","root.basket[0].color[0]","=","Red")',
                'binary("!count(root.basket[0])=1||root.basket[0].color[0]=Red","!count(root.basket[0])=1","||","root.basket[0].color[0]=Red")',
                'unary("!count(root.basket[0])=1","!","count(root.basket[0])=1")',
                'binary("count(root.basket[0])=1","count(root.basket[0])","=","1")',
                'function("count(root.basket[0])","count","root.basket[0]")',
                'set("root.basket[0]","root.basket[0]")',
                'number("1",1)',
            }
        ),
    },
    "require_with_partonomy2": {
        "test": StableModels(
            {
                'constant("Red")',
                'discrete("Color")',
                'part("product")',
                'part("Bag")',
                'constraint(("root.bag",2),"lowerbound")',
                'constraint(("root.bag[0].color",1),"lowerbound")',
                'constraint(("root.bag[1].color",1),"lowerbound")',
                'constraint((0,"!count(root.bag[0])=1||root.bag[0].color[0]=Red"),"boolean")',
                'constraint((0,"!count(root.bag[1])=1||root.bag[1].color[0]=Red"),"boolean")',
                'domain("Color","Red")',
                'domain("Color","Yellow")',
                'index("root.bag[0]",0)',
                'index("root.bag[1]",1)',
                'index("root.bag[1].color[0]",0)',
                'index("root.bag[0].color[0]",0)',
                'number("1",1)',
                'parent("root.bag[0]","root")',
                'parent("root.bag[1]","root")',
                'parent("root.bag[1].color[0]","root.bag[1]")',
                'parent("root.bag[0].color[0]","root.bag[0]")',
                'set("root.bag","root.bag[0]")',
                'set("root.bag","root.bag[1]")',
                'set("root.bag[0].color","root.bag[0].color[0]")',
                'set("root.bag[1].color","root.bag[1].color[0]")',
                'set("root.bag[0]","root.bag[0]")',
                'set("root.bag[1]","root.bag[1]")',
                'type("root","product")',
                'type("root.bag[0]","Bag")',
                'type("root.bag[1]","Bag")',
                'type("root.bag[1].color[0]","Color")',
                'type("root.bag[0].color[0]","Color")',
                'function("count(root.bag[0])","count","root.bag[0]")',
                'function("count(root.bag[1])","count","root.bag[1]")',
                'unary("!count(root.bag[0])=1","!","count(root.bag[0])=1")',
                'unary("!count(root.bag[1])=1","!","count(root.bag[1])=1")',
                'binary("root.bag[1].color[0]=Red","root.bag[1].color[0]","=","Red")',
                'binary("root.bag[0].color[0]=Red","root.bag[0].color[0]","=","Red")',
                'binary("count(root.bag[0])=1","count(root.bag[0])","=","1")',
                'binary("count(root.bag[1])=1","count(root.bag[1])","=","1")',
                'binary("!count(root.bag[0])=1||root.bag[0].color[0]=Red","!count(root.bag[0])=1","||","root.bag[0].color[0]=Red")',
                'binary("!count(root.bag[1])=1||root.bag[1].color[0]=Red","!count(root.bag[1])=1","||","root.bag[1].color[0]=Red")',
            }
        ),
    },
    "require_with_partonomy_multiple_instances": {
        "test": StableModels(
            {
                'constant("Red")',
                'discrete("Color")',
                'part("product")',
                'part("Compartment")',
                'part("Bag")',
                'constraint(("root.compartment",2),"lowerbound")',
                'constraint(("root.compartment[0].bag",2),"lowerbound")',
                'constraint(("root.compartment[1].bag",2),"lowerbound")',
                'constraint(("root.compartment[1].bag[0].color",1),"lowerbound")',
                'constraint(("root.compartment[1].bag[1].color",1),"lowerbound")',
                'constraint(("root.compartment[0].bag[0].color",1),"lowerbound")',
                'constraint(("root.compartment[0].bag[1].color",1),"lowerbound")',
                'constraint((0,"!count(root.compartment[0])=1||root.compartment[0].bag[0].color[0]=Red"),"boolean")',
                'constraint((0,"!count(root.compartment[0])=1||root.compartment[0].bag[1].color[0]=Red"),"boolean")',
                'constraint((0,"!count(root.compartment[1])=1||root.compartment[1].bag[0].color[0]=Red"),"boolean")',
                'constraint((0,"!count(root.compartment[1])=1||root.compartment[1].bag[1].color[0]=Red"),"boolean")',
                'domain("Color","Red")',
                'domain("Color","Yellow")',
                'domain("Color","Green")',
                'index("root.compartment[0]",0)',
                'index("root.compartment[1]",1)',
                'index("root.compartment[1].bag[0]",0)',
                'index("root.compartment[1].bag[1]",1)',
                'index("root.compartment[0].bag[0]",0)',
                'index("root.compartment[0].bag[1]",1)',
                'index("root.compartment[0].bag[1].color[0]",0)',
                'index("root.compartment[0].bag[0].color[0]",0)',
                'index("root.compartment[1].bag[1].color[0]",0)',
                'index("root.compartment[1].bag[0].color[0]",0)',
                'number("1",1)',
                'parent("root.compartment[0]","root")',
                'parent("root.compartment[1]","root")',
                'parent("root.compartment[1].bag[0]","root.compartment[1]")',
                'parent("root.compartment[1].bag[1]","root.compartment[1]")',
                'parent("root.compartment[0].bag[0]","root.compartment[0]")',
                'parent("root.compartment[0].bag[1]","root.compartment[0]")',
                'parent("root.compartment[0].bag[1].color[0]","root.compartment[0].bag[1]")',
                'parent("root.compartment[0].bag[0].color[0]","root.compartment[0].bag[0]")',
                'parent("root.compartment[1].bag[1].color[0]","root.compartment[1].bag[1]")',
                'parent("root.compartment[1].bag[0].color[0]","root.compartment[1].bag[0]")',
                'set("root.compartment","root.compartment[0]")',
                'set("root.compartment","root.compartment[1]")',
                'set("root.compartment[0].bag","root.compartment[0].bag[0]")',
                'set("root.compartment[0].bag","root.compartment[0].bag[1]")',
                'set("root.compartment[1].bag","root.compartment[1].bag[0]")',
                'set("root.compartment[1].bag","root.compartment[1].bag[1]")',
                'set("root.compartment[1].bag[0].color","root.compartment[1].bag[0].color[0]")',
                'set("root.compartment[1].bag[1].color","root.compartment[1].bag[1].color[0]")',
                'set("root.compartment[0].bag[0].color","root.compartment[0].bag[0].color[0]")',
                'set("root.compartment[0].bag[1].color","root.compartment[0].bag[1].color[0]")',
                'set("root.compartment[0]","root.compartment[0]")',
                'set("root.compartment[1]","root.compartment[1]")',
                'type("root","product")',
                'type("root.compartment[0]","Compartment")',
                'type("root.compartment[1]","Compartment")',
                'type("root.compartment[1].bag[0]","Bag")',
                'type("root.compartment[1].bag[1]","Bag")',
                'type("root.compartment[0].bag[0]","Bag")',
                'type("root.compartment[0].bag[1]","Bag")',
                'type("root.compartment[0].bag[1].color[0]","Color")',
                'type("root.compartment[0].bag[0].color[0]","Color")',
                'type("root.compartment[1].bag[1].color[0]","Color")',
                'type("root.compartment[1].bag[0].color[0]","Color")',
                'function("count(root.compartment[0])","count","root.compartment[0]")',
                'function("count(root.compartment[1])","count","root.compartment[1]")',
                'unary("!count(root.compartment[0])=1","!","count(root.compartment[0])=1")',
                'unary("!count(root.compartment[1])=1","!","count(root.compartment[1])=1")',
                'binary("root.compartment[1].bag[1].color[0]=Red","root.compartment[1].bag[1].color[0]","=","Red")',
                'binary("root.compartment[1].bag[0].color[0]=Red","root.compartment[1].bag[0].color[0]","=","Red")',
                'binary("root.compartment[0].bag[1].color[0]=Red","root.compartment[0].bag[1].color[0]","=","Red")',
                'binary("root.compartment[0].bag[0].color[0]=Red","root.compartment[0].bag[0].color[0]","=","Red")',
                'binary("count(root.compartment[0])=1","count(root.compartment[0])","=","1")',
                'binary("count(root.compartment[1])=1","count(root.compartment[1])","=","1")',
                'binary("!count(root.compartment[0])=1||root.compartment[0].bag[0].color[0]=Red","!count(root.compartment[0])=1","||","root.compartment[0].bag[0].color[0]=Red")',
                'binary("!count(root.compartment[0])=1||root.compartment[0].bag[1].color[0]=Red","!count(root.compartment[0])=1","||","root.compartment[0].bag[1].color[0]=Red")',
                'binary("!count(root.compartment[1])=1||root.compartment[1].bag[0].color[0]=Red","!count(root.compartment[1])=1","||","root.compartment[1].bag[0].color[0]=Red")',
                'binary("!count(root.compartment[1])=1||root.compartment[1].bag[1].color[0]=Red","!count(root.compartment[1])=1","||","root.compartment[1].bag[1].color[0]=Red")',
            }
        ),
    },
    "combination": {
        "test": StableModels(
            {
                'discrete("Wheel")',
                'discrete("Bool")',
                'part("product")',
                'constraint(("root.wheelSupport",1),"lowerbound")',
                'constraint(("root.wheel",1),"lowerbound")',
                'constraint((0,"root"),"table")',
                'domain("Wheel","W14")',
                'domain("Wheel","W16")',
                'domain("Wheel","W18")',
                'domain("Wheel","W20")',
                'domain("Bool","True")',
                'domain("Bool","False")',
                'index("root.wheelSupport[0]",0)',
                'index("root.wheel[0]",0)',
                'parent("root.wheelSupport[0]","root")',
                'parent("root.wheel[0]","root")',
                'set("root.wheelSupport","root.wheelSupport[0]")',
                'set("root.wheel","root.wheel[0]")',
                'type("root","product")',
                'type("root.wheelSupport[0]","Bool")',
                'type("root.wheel[0]","Wheel")',
                'allow(0,(0,0),"True")',
                'allow(0,(1,0),"W14")',
                'allow(0,(1,0),"W16")',
                'allow(0,(0,1),"False")',
                'allow(0,(1,1),"W18")',
                'allow(0,(1,1),"W20")',
                'column((0,"root"),("root.wheelSupport[0]",("root.wheel[0]",())),0,"root.wheelSupport[0]")',
                'column((0,"root"),("root.wheelSupport[0]",("root.wheel[0]",())),1,"root.wheel[0]")',
            }
        ),
    },
    "combination_with_structure": {
        "test": StableModels(
            {
                'discrete("Size")',
                'discrete("Bool")',
                'part("product")',
                'part("Wheel")',
                'constraint(("root.wheelSupport",1),"lowerbound")',
                'constraint(("root.wheel",2),"lowerbound")',
                'constraint(("root.wheel[0].size",1),"lowerbound")',
                'constraint(("root.wheel[1].size",1),"lowerbound")',
                'constraint((0,"root"),"table")',
                'domain("Size","W14")',
                'domain("Size","W16")',
                'domain("Size","W18")',
                'domain("Size","W20")',
                'domain("Bool","True")',
                'domain("Bool","False")',
                'index("root.wheelSupport[0]",0)',
                'index("root.wheel[0]",0)',
                'index("root.wheel[1]",1)',
                'index("root.wheel[1].size[0]",0)',
                'index("root.wheel[0].size[0]",0)',
                'parent("root.wheelSupport[0]","root")',
                'parent("root.wheel[0]","root")',
                'parent("root.wheel[1]","root")',
                'parent("root.wheel[1].size[0]","root.wheel[1]")',
                'parent("root.wheel[0].size[0]","root.wheel[0]")',
                'set("root.wheelSupport","root.wheelSupport[0]")',
                'set("root.wheel","root.wheel[0]")',
                'set("root.wheel","root.wheel[1]")',
                'set("root.wheel[0].size","root.wheel[0].size[0]")',
                'set("root.wheel[1].size","root.wheel[1].size[0]")',
                'type("root","product")',
                'type("root.wheelSupport[0]","Bool")',
                'type("root.wheel[0]","Wheel")',
                'type("root.wheel[1]","Wheel")',
                'type("root.wheel[1].size[0]","Size")',
                'type("root.wheel[0].size[0]","Size")',
                'allow(0,(0,0),"True")',
                'allow(0,(1,0),"W14")',
                'allow(0,(1,0),"W16")',
                'allow(0,(0,1),"False")',
                'allow(0,(1,1),"W18")',
                'allow(0,(1,1),"W20")',
                'column((0,"root"),("root.wheelSupport[0]",("root.wheel[0].size[0]",())),0,"root.wheelSupport[0]")',
                'column((0,"root"),("root.wheelSupport[0]",("root.wheel[1].size[0]",())),0,"root.wheelSupport[0]")',
                'column((0,"root"),("root.wheelSupport[0]",("root.wheel[1].size[0]",())),1,"root.wheel[1].size[0]")',
                'column((0,"root"),("root.wheelSupport[0]",("root.wheel[0].size[0]",())),1,"root.wheel[0].size[0]")',
            }
        ),
    },
    "combination_at_part_with_wildcard": {
        "test": StableModels(
            {
                'discrete("Size")',
                'discrete("Material")',
                'part("product")',
                'part("Wheel")',
                'constraint(("root.wheel",2),"lowerbound")',
                'constraint(("root.wheel[0].size",1),"lowerbound")',
                'constraint(("root.wheel[1].size",1),"lowerbound")',
                'constraint(("root.wheel[0].material",1),"lowerbound")',
                'constraint(("root.wheel[1].material",1),"lowerbound")',
                'constraint((0,"root.wheel[0]"),"table")',
                'constraint((0,"root.wheel[1]"),"table")',
                'constraint((1,"root.wheel[0].size[0]=root.wheel[1].size[0]"),"boolean")',
                'constraint((1,"root.wheel[1].size[0]=root.wheel[1].size[0]"),"boolean")',
                'constraint((1,"root.wheel[0].size[0]=root.wheel[0].size[0]"),"boolean")',
                'constraint((1,"root.wheel[1].size[0]=root.wheel[0].size[0]"),"boolean")',
                'domain("Size","W28")',
                'domain("Size","W30")',
                'domain("Material","Carbon")',
                'domain("Material","Aluminum")',
                'index("root.wheel[0]",0)',
                'index("root.wheel[1]",1)',
                'index("root.wheel[1].size[0]",0)',
                'index("root.wheel[1].material[0]",0)',
                'index("root.wheel[0].size[0]",0)',
                'index("root.wheel[0].material[0]",0)',
                'parent("root.wheel[0]","root")',
                'parent("root.wheel[1]","root")',
                'parent("root.wheel[1].size[0]","root.wheel[1]")',
                'parent("root.wheel[1].material[0]","root.wheel[1]")',
                'parent("root.wheel[0].size[0]","root.wheel[0]")',
                'parent("root.wheel[0].material[0]","root.wheel[0]")',
                'set("root.wheel","root.wheel[0]")',
                'set("root.wheel","root.wheel[1]")',
                'set("root.wheel[0].size","root.wheel[0].size[0]")',
                'set("root.wheel[1].size","root.wheel[1].size[0]")',
                'set("root.wheel[0].material","root.wheel[0].material[0]")',
                'set("root.wheel[1].material","root.wheel[1].material[0]")',
                'type("root","product")',
                'type("root.wheel[0]","Wheel")',
                'type("root.wheel[1]","Wheel")',
                'type("root.wheel[1].size[0]","Size")',
                'type("root.wheel[1].material[0]","Material")',
                'type("root.wheel[0].size[0]","Size")',
                'type("root.wheel[0].material[0]","Material")',
                'allow(0,(0,0),"W28")',
                'allow(0,(0,1),"W30")',
                'allow(0,(1,1),"Aluminum")',
                'binary("root.wheel[1].size[0]=root.wheel[0].size[0]","root.wheel[1].size[0]","=","root.wheel[0].size[0]")',
                'binary("root.wheel[0].size[0]=root.wheel[0].size[0]","root.wheel[0].size[0]","=","root.wheel[0].size[0]")',
                'binary("root.wheel[1].size[0]=root.wheel[1].size[0]","root.wheel[1].size[0]","=","root.wheel[1].size[0]")',
                'binary("root.wheel[0].size[0]=root.wheel[1].size[0]","root.wheel[0].size[0]","=","root.wheel[1].size[0]")',
                'column((0,"root.wheel[1]"),("root.wheel[1].size[0]",("root.wheel[1].material[0]",())),0,"root.wheel[1].size[0]")',
                'column((0,"root.wheel[0]"),("root.wheel[0].size[0]",("root.wheel[0].material[0]",())),0,"root.wheel[0].size[0]")',
                'column((0,"root.wheel[1]"),("root.wheel[1].size[0]",("root.wheel[1].material[0]",())),1,"root.wheel[1].material[0]")',
                'column((0,"root.wheel[0]"),("root.wheel[0].size[0]",("root.wheel[0].material[0]",())),1,"root.wheel[0].material[0]")',
            }
        ),
    },
    "combination_at_part_multiple_instances": {
        "test": StableModels(
            {
                'discrete("Wheel")',
                'discrete("Material")',
                'part("product")',
                'part("Bike")',
                'constraint(("root.bike",2),"lowerbound")',
                'constraint(("root.bike[0].wheel",2),"lowerbound")',
                'constraint(("root.bike[1].wheel",2),"lowerbound")',
                'constraint(("root.bike[0].material",1),"lowerbound")',
                'constraint(("root.bike[1].material",1),"lowerbound")',
                'constraint((0,"root.bike[0]"),"table")',
                'constraint((0,"root.bike[1]"),"table")',
                # 'constraint((1,"root.bike[1].wheel[1]=root.bike[1].wheel[0]"),"boolean")',
                # 'constraint((1,"root.bike[1].wheel[0]=root.bike[1].wheel[0]"),"boolean")',
                # 'constraint((1,"root.bike[1].wheel[1]=root.bike[1].wheel[1]"),"boolean")',
                # 'constraint((1,"root.bike[1].wheel[0]=root.bike[1].wheel[1]"),"boolean")',
                # 'constraint((1,"root.bike[0].wheel[1]=root.bike[0].wheel[0]"),"boolean")',
                # 'constraint((1,"root.bike[0].wheel[0]=root.bike[0].wheel[0]"),"boolean")',
                # 'constraint((1,"root.bike[0].wheel[1]=root.bike[0].wheel[1]"),"boolean")',
                # 'constraint((1,"root.bike[0].wheel[0]=root.bike[0].wheel[1]"),"boolean")',
                'domain("Wheel","W28")',
                'domain("Wheel","W30")',
                'domain("Material","Carbon")',
                'domain("Material","Aluminum")',
                'index("root.bike[0]",0)',
                'index("root.bike[1]",1)',
                'index("root.bike[1].wheel[0]",0)',
                'index("root.bike[1].wheel[1]",1)',
                'index("root.bike[1].material[0]",0)',
                'index("root.bike[0].wheel[0]",0)',
                'index("root.bike[0].wheel[1]",1)',
                'index("root.bike[0].material[0]",0)',
                'parent("root.bike[0]","root")',
                'parent("root.bike[1]","root")',
                'parent("root.bike[1].wheel[0]","root.bike[1]")',
                'parent("root.bike[1].wheel[1]","root.bike[1]")',
                'parent("root.bike[1].material[0]","root.bike[1]")',
                'parent("root.bike[0].wheel[0]","root.bike[0]")',
                'parent("root.bike[0].wheel[1]","root.bike[0]")',
                'parent("root.bike[0].material[0]","root.bike[0]")',
                'set("root.bike","root.bike[0]")',
                'set("root.bike","root.bike[1]")',
                'set("root.bike[0].wheel","root.bike[0].wheel[0]")',
                'set("root.bike[0].wheel","root.bike[0].wheel[1]")',
                'set("root.bike[1].wheel","root.bike[1].wheel[0]")',
                'set("root.bike[1].wheel","root.bike[1].wheel[1]")',
                'set("root.bike[0].material","root.bike[0].material[0]")',
                'set("root.bike[1].material","root.bike[1].material[0]")',
                'type("root","product")',
                'type("root.bike[0]","Bike")',
                'type("root.bike[1]","Bike")',
                'type("root.bike[1].wheel[0]","Wheel")',
                'type("root.bike[1].wheel[1]","Wheel")',
                'type("root.bike[1].material[0]","Material")',
                'type("root.bike[0].wheel[0]","Wheel")',
                'type("root.bike[0].wheel[1]","Wheel")',
                'type("root.bike[0].material[0]","Material")',
                'allow(0,(0,0),"W28")',
                'allow(0,(1,0),"Carbon")',
                'allow(0,(0,1),"W30")',
                'allow(0,(1,1),"Aluminum")',
                # 'binary("root.bike[0].wheel[0]=root.bike[0].wheel[1]","root.bike[0].wheel[0]","=","root.bike[0].wheel[1]")',
                # 'binary("root.bike[0].wheel[1]=root.bike[0].wheel[1]","root.bike[0].wheel[1]","=","root.bike[0].wheel[1]")',
                # 'binary("root.bike[0].wheel[0]=root.bike[0].wheel[0]","root.bike[0].wheel[0]","=","root.bike[0].wheel[0]")',
                # 'binary("root.bike[0].wheel[1]=root.bike[0].wheel[0]","root.bike[0].wheel[1]","=","root.bike[0].wheel[0]")',
                # 'binary("root.bike[1].wheel[0]=root.bike[1].wheel[1]","root.bike[1].wheel[0]","=","root.bike[1].wheel[1]")',
                # 'binary("root.bike[1].wheel[1]=root.bike[1].wheel[1]","root.bike[1].wheel[1]","=","root.bike[1].wheel[1]")',
                # 'binary("root.bike[1].wheel[0]=root.bike[1].wheel[0]","root.bike[1].wheel[0]","=","root.bike[1].wheel[0]")',
                # 'binary("root.bike[1].wheel[1]=root.bike[1].wheel[0]","root.bike[1].wheel[1]","=","root.bike[1].wheel[0]")',
                'column((0,"root.bike[1]"),("root.bike[1].wheel[0]",("root.bike[1].material[0]",())),0,"root.bike[1].wheel[0]")',
                'column((0,"root.bike[1]"),("root.bike[1].wheel[1]",("root.bike[1].material[0]",())),0,"root.bike[1].wheel[1]")',
                'column((0,"root.bike[0]"),("root.bike[0].wheel[0]",("root.bike[0].material[0]",())),0,"root.bike[0].wheel[0]")',
                'column((0,"root.bike[0]"),("root.bike[0].wheel[1]",("root.bike[0].material[0]",())),0,"root.bike[0].wheel[1]")',
                'column((0,"root.bike[1]"),("root.bike[1].wheel[1]",("root.bike[1].material[0]",())),1,"root.bike[1].material[0]")',
                'column((0,"root.bike[1]"),("root.bike[1].wheel[0]",("root.bike[1].material[0]",())),1,"root.bike[1].material[0]")',
                'column((0,"root.bike[0]"),("root.bike[0].wheel[1]",("root.bike[0].material[0]",())),1,"root.bike[0].material[0]")',
                'column((0,"root.bike[0]"),("root.bike[0].wheel[0]",("root.bike[0].material[0]",())),1,"root.bike[0].material[0]")',
            }
        ),
    },
    "simple_numeric_feature": {
        "test": StableModels(
            {
                'integer("product.size")',
                'part("product")',
                'constraint(("root.size",1),"lowerbound")',
                'index("root.size[0]",0)',
                'parent("root.size[0]","root")',
                'set("root.size","root.size[0]")',
                'type("root","product")',
                'type("root.size[0]","product.size")',
                'range("product.size",1,3)',
            }
        ),
        "program": """
            coom_structure("product").
            coom_feature("product","size","num",1,1).
            coom_range("product","size",1,3).""",
    },
    "simple_arithmetic_plus": {
        "test": StableModels(
            {
                'integer("product.a")',
                'integer("product.b")',
                'part("product")',
                'constraint(("root.a",1),"lowerbound")',
                'constraint(("root.b",1),"lowerbound")',
                'constraint((0,"root.a[0]+root.b[0]<6"),"boolean")',
                'index("root.a[0]",0)',
                'index("root.b[0]",0)',
                'number("6",6)',
                'parent("root.a[0]","root")',
                'parent("root.b[0]","root")',
                'set("root.a","root.a[0]")',
                'set("root.b","root.b[0]")',
                'type("root","product")',
                'type("root.a[0]","product.a")',
                'type("root.b[0]","product.b")',
                'range("product.a",1,1)',
                'range("product.b",3,3)',
                'binary("root.a[0]+root.b[0]","root.a[0]","+","root.b[0]")',
                'binary("root.a[0]+root.b[0]<6","root.a[0]+root.b[0]","<","6")',
            }
        ),
    },
    "simple_arithmetic_minus": {
        "test": StableModels(
            {
                'integer("product.a")',
                'integer("product.b")',
                'part("product")',
                'constraint(("root.a",1),"lowerbound")',
                'constraint(("root.b",1),"lowerbound")',
                'constraint((0,"root.b[0]-root.a[0]>=3"),"boolean")',
                'index("root.a[0]",0)',
                'index("root.b[0]",0)',
                'number("3",3)',
                'parent("root.a[0]","root")',
                'parent("root.b[0]","root")',
                'set("root.a","root.a[0]")',
                'set("root.b","root.b[0]")',
                'type("root","product")',
                'type("root.a[0]","product.a")',
                'type("root.b[0]","product.b")',
                'range("product.a",1,3)',
                'range("product.b",3,4)',
                'binary("root.b[0]-root.a[0]","root.b[0]","-","root.a[0]")',
                'binary("root.b[0]-root.a[0]>=3","root.b[0]-root.a[0]",">=","3")',
            }
        ),
    },
    "simple_arithmetic_multiplication": {
        "test": StableModels(
            {
                'integer("product.a")',
                'integer("product.b")',
                'part("product")',
                'constraint(("root.a",1),"lowerbound")',
                'constraint(("root.b",1),"lowerbound")',
                'constraint((0,"root.a[0]*root.b[0]>=10"),"boolean")',
                'index("root.a[0]",0)',
                'index("root.b[0]",0)',
                'number("10",10)',
                'parent("root.a[0]","root")',
                'parent("root.b[0]","root")',
                'set("root.a","root.a[0]")',
                'set("root.b","root.b[0]")',
                'type("root","product")',
                'type("root.a[0]","product.a")',
                'type("root.b[0]","product.b")',
                'range("product.a",1,3)',
                'range("product.b",3,4)',
                'binary("root.a[0]*root.b[0]","root.a[0]","*","root.b[0]")',
                'binary("root.a[0]*root.b[0]>=10","root.a[0]*root.b[0]",">=","10")',
            }
        ),
    },
    "simple_arithmetic_plus_default_right": {
        "test": StableModels(
            {
                'integer("product.a")',
                'part("product")',
                'constraint(("root.a",1),"lowerbound")',
                'index("root.a[0]",0)',
                'number("2",2)',
                'parent("root.a[0]","root")',
                'set("root.a","root.a[0]")',
                'type("root","product")',
                'type("root.a[0]","product.a")',
                'range("product.a",1,3)',
                'binary("root.a[0]=2","root.a[0]","=","2")',
                'constraint((0,"root.a[0]=2"),"boolean")',
            }
        ),
    },
    "simple_arithmetic_plus_default_left": {
        "test": StableModels(
            {
                'integer("product.b")',
                'part("product")',
                'constraint(("root.b",1),"lowerbound")',
                'index("root.b[0]",0)',
                'number("2",2)',
                'parent("root.b[0]","root")',
                'set("root.b","root.b[0]")',
                'type("root","product")',
                'type("root.b[0]","product.b")',
                'range("product.b",1,3)',
                'binary("root.b[0]=2","root.b[0]","=","2")',
                'constraint((0,"root.b[0]=2"),"boolean")',
            }
        ),
    },
    "simple_arithmetic_minus_default_right": {
        "test": StableModels(
            {
                'integer("product.a")',
                'part("product")',
                'constraint(("root.a",1),"lowerbound")',
                'index("root.a[0]",0)',
                'number("2",2)',
                'parent("root.a[0]","root")',
                'set("root.a","root.a[0]")',
                'type("root","product")',
                'type("root.a[0]","product.a")',
                'range("product.a",1,3)',
            }
        ),
    },
    "simple_arithmetic_minus_default_left": {
        "test": StableModels(
            {
                'integer("product.b")',
                'part("product")',
                'constraint(("root.b",1),"lowerbound")',
                'index("root.b[0]",0)',
                'number("2",2)',
                'parent("root.b[0]","root")',
                'set("root.b","root.b[0]")',
                'type("root","product")',
                'type("root.b[0]","product.b")',
                'range("product.b",1,3)',
            }
        ),
    },
    "parentheses": {
        "test": StableModels(
            {
                'integer("product.a")',
                'integer("product.b")',
                'part("product")',
                'constraint(("root.a",1),"lowerbound")',
                'constraint(("root.b",1),"lowerbound")',
                'constraint((0,"root.a[0]=(root.b[0])"),"boolean")',
                'index("root.a[0]",0)',
                'index("root.b[0]",0)',
                'parent("root.a[0]","root")',
                'parent("root.b[0]","root")',
                'set("root.a","root.a[0]")',
                'set("root.b","root.b[0]")',
                'type("root","product")',
                'type("root.a[0]","product.a")',
                'type("root.b[0]","product.b")',
                'range("product.a",1,2)',
                'range("product.b",1,3)',
                'unary("(root.b[0])","()","root.b[0]")',
                'binary("root.a[0]=(root.b[0])","root.a[0]","=","(root.b[0])")',
            }
        ),
    },
    "count": {
        "test": StableModels(
            {
                'part("product")',
                'part("Bag")',
                'constraint(("root.bag",0),"lowerbound")',
                'constraint((0,"count(root.bag)<=2"),"boolean")',
                'index("root.bag[0]",0)',
                'index("root.bag[1]",1)',
                'index("root.bag[2]",2)',
                'number("2",2)',
                'parent("root.bag[0]","root")',
                'parent("root.bag[1]","root")',
                'parent("root.bag[2]","root")',
                'set("root.bag","root.bag[0]")',
                'set("root.bag","root.bag[1]")',
                'set("root.bag","root.bag[2]")',
                'type("root","product")',
                'type("root.bag[0]","Bag")',
                'type("root.bag[1]","Bag")',
                'type("root.bag[2]","Bag")',
                'function("count(root.bag)","count","root.bag")',
                'binary("count(root.bag)<=2","count(root.bag)","<=","2")',
            }
        ),
    },
    "sum": {
        "test": StableModels(
            {
                'integer("product.x")',
                'part("product")',
                'constraint(("root.x",0),"lowerbound")',
                'constraint((0,"sum(root.x)<=10"),"boolean")',
                'index("root.x[0]",0)',
                'index("root.x[1]",1)',
                'index("root.x[2]",2)',
                'number("10",10)',
                'parent("root.x[0]","root")',
                'parent("root.x[1]","root")',
                'parent("root.x[2]","root")',
                'set("root.x","root.x[0]")',
                'set("root.x","root.x[1]")',
                'set("root.x","root.x[2]")',
                'type("root","product")',
                'type("root.x[0]","product.x")',
                'type("root.x[1]","product.x")',
                'type("root.x[2]","product.x")',
                'function("sum(root.x)","sum","root.x")',
                'range("product.x",0,10)',
                'binary("sum(root.x)<=10","sum(root.x)","<=","10")',
            }
        ),
    },
    "simple_default": {
        "test": StableModels(
            {
                'discrete("Color")',
                'domain("Color","Red")',
                'domain("Color","White")',
                'domain("Color","Black")',
                'type("root","product")',
                'type("root.color[0]","Color")',
                'index("root.color[0]",0)',
                'parent("root.color[0]","root")',
                'constraint(("root.color",1),"lowerbound")',
                'set("root.color","root.color[0]")',
                'part("product")',
                'constant("White")',
                'default("root.color[0]","White")',
            }
        ),
    },
    "minimize": {
        "test": StableModels(
            {
                'integer("product.totalWeight")',
                'range("product.totalWeight",1,10)',
                'type("root","product")',
                'type("root.totalWeight[0]","product.totalWeight")',
                'index("root.totalWeight[0]",0)',
                'parent("root.totalWeight[0]","root")',
                'constraint(("root.totalWeight",1),"lowerbound")',
                'set("root.totalWeight","root.totalWeight[0]")',
                'part("product")',
                'minimize("root.totalWeight[0]",0)',
            }
        ),
    },
    "maximize": {
        "test": StableModels(
            {
                'integer("product.totalOutput")',
                'range("product.totalOutput",1,10)',
                'type("root","product")',
                'type("root.totalOutput[0]","product.totalOutput")',
                'index("root.totalOutput[0]",0)',
                'parent("root.totalOutput[0]","root")',
                'constraint(("root.totalOutput",1),"lowerbound")',
                'set("root.totalOutput","root.totalOutput[0]")',
                'part("product")',
                'maximize("root.totalOutput[0]",1)',
            }
        ),
    },
    "minimize_non_root_path": {
        "test": StableModels(
            {
                'integer("Bag.weight")',
                'range("Bag.weight",0,10)',
                'type("root","product")',
                'type("root.bags[0]","Bag")',
                'type("root.bags[0].weight[0]","Bag.weight")',
                'index("root.bags[0]",0)',
                'index("root.bags[0].weight[0]",0)',
                'parent("root.bags[0]","root")',
                'parent("root.bags[0].weight[0]","root.bags[0]")',
                'constraint(("root.bags",1),"lowerbound")',
                'constraint(("root.bags[0].weight",1),"lowerbound")',
                'set("root.bags","root.bags[0]")',
                'set("root.bags[0].weight","root.bags[0].weight[0]")',
                'part("product")',
                'part("Bag")',
                'minimize("root.bags[0].weight[0]",1)',
            }
        ),
    },
    "maximize_non_root_path": {
        "test": StableModels(
            {
                'integer("Bag.volume")',
                'range("Bag.volume",0,10)',
                'type("root","product")',
                'type("root.bags[0]","Bag")',
                'type("root.bags[0].volume[0]","Bag.volume")',
                'index("root.bags[0]",0)',
                'index("root.bags[0].volume[0]",0)',
                'parent("root.bags[0]","root")',
                'parent("root.bags[0].volume[0]","root.bags[0]")',
                'constraint(("root.bags",1),"lowerbound")',
                'constraint(("root.bags[0].volume",1),"lowerbound")',
                'set("root.bags","root.bags[0]")',
                'set("root.bags[0].volume","root.bags[0].volume[0]")',
                'part("product")',
                'part("Bag")',
                'minimize("root.bags[0].volume[0]",1)',
            }
        ),
    },
    "minimize_function": {
        "test": StableModels(
            {
                'integer("Bag.volume")',
                'range("Bag.volume",0,10)',
                'type("root","product")',
                'type("root.bags[0]","Bag")',
                'type("root.bags[1]","Bag")',
                'type("root.bags[1].volume[0]","Bag.volume")',
                'type("root.bags[0].volume[0]","Bag.volume")',
                'index("root.bags[0]",0)',
                'index("root.bags[1]",1)',
                'index("root.bags[1].volume[0]",0)',
                'index("root.bags[0].volume[0]",0)',
                'parent("root.bags[0]","root")',
                'parent("root.bags[1]","root")',
                'parent("root.bags[1].volume[0]","root.bags[1]")',
                'parent("root.bags[0].volume[0]","root.bags[0]")',
                'constraint(("root.bags",0),"lowerbound")',
                'constraint(("root.bags[0].volume",1),"lowerbound")',
                'constraint(("root.bags[1].volume",1),"lowerbound")',
                'function("count(root.bags)","count","root.bags")',
                'set("root.bags","root.bags[0]")',
                'set("root.bags","root.bags[1]")',
                'set("root.bags[0].volume","root.bags[0].volume[0]")',
                'set("root.bags[1].volume","root.bags[1].volume[0]")',
                'part("product")',
                'part("Bag")',
                'minimize("count(root.bags)",0)',
            }
        ),
    },
    "maximize_function": {
        "test": StableModels(
            {
                'integer("Bag.volume")',
                'range("Bag.volume",0,10)',
                'type("root","product")',
                'type("root.bags[0]","Bag")',
                'type("root.bags[1]","Bag")',
                'type("root.bags[1].volume[0]","Bag.volume")',
                'type("root.bags[0].volume[0]","Bag.volume")',
                'index("root.bags[0]",0)',
                'index("root.bags[1]",1)',
                'index("root.bags[1].volume[0]",0)',
                'index("root.bags[0].volume[0]",0)',
                'parent("root.bags[0]","root")',
                'parent("root.bags[1]","root")',
                'parent("root.bags[1].volume[0]","root.bags[1]")',
                'parent("root.bags[0].volume[0]","root.bags[0]")',
                'constraint(("root.bags",0),"lowerbound")',
                'constraint(("root.bags[0].volume",1),"lowerbound")',
                'constraint(("root.bags[1].volume",1),"lowerbound")',
                'function("sum(root.bags.volume)","sum","root.bags.volume")',
                'set("root.bags","root.bags[0]")',
                'set("root.bags","root.bags[1]")',
                'set("root.bags[0].volume","root.bags[0].volume[0]")',
                'set("root.bags[1].volume","root.bags[1].volume[0]")',
                'set("root.bags.volume","root.bags[1].volume[0]")',
                'set("root.bags.volume","root.bags[0].volume[0]")',
                'part("product")',
                'part("Bag")',
                'maximize("sum(root.bags.volume)",0)',
            }
        ),
    },
    "set_constant": {
        "test": StableModels({'user_value("root.color[0]","Yellow")'}),
        "program": 'coom_user_value("root.color[0]","Yellow").',
    },
    "set_number": {
        "test": StableModels({'user_value("root.size[0]",5)'}),
        "program": 'coom_user_value("root.size[0]",5).',
    },
    "add": {
        "test": StableModels({'user_include("root.bag[0]")'}),
        "program": 'coom_user_include("root.bag[0]").',
    },
    "unbounded_singleshot_zero_lb_zero_max": {
        "test": StableModels(get_model_from_file("unbounded_singleshot_zero_lb_zero_max.lp")),
        "files": ["unbounded_zero_lb.lp"],
    },
    "unbounded_singleshot_zero_lb_one_max": {
        "test": StableModels(get_model_from_file("unbounded_singleshot_zero_lb_one_max.lp")),
        "files": ["unbounded_zero_lb.lp"],
    },
    "unbounded_singleshot_two_lb_zero_max": {
        "test": StableModels(get_model_from_file("unbounded_singleshot_two_lb_zero_max.lp")),
        "files": ["unbounded_two_lb.lp"],
    },
    "unbounded_singleshot_two_lb_one_max": {
        "test": StableModels(get_model_from_file("unbounded_singleshot_two_lb_one_max.lp")),
        "files": ["unbounded_two_lb.lp"],
    },
    "unbounded_singleshot_optimize": {
        "test": StableModels(get_model_from_file("unbounded_singleshot_optimize.lp")),
        "files": ["unbounded_optimize.lp"],
    },
    "unbounded_multishot_zero_lb_zero_max_superset": {
        "test": Supersets(get_model_from_file("unbounded_singleshot_zero_lb_zero_max.lp")),
        "files": ["unbounded_zero_lb.lp"],
    },
    "unbounded_multishot_zero_lb_one_max_superset": {
        "test": Supersets(get_model_from_file("unbounded_singleshot_zero_lb_one_max.lp")),
        "files": ["unbounded_zero_lb.lp"],
    },
    "unbounded_multishot_two_lb_zero_max_superset": {
        "test": Supersets(get_model_from_file("unbounded_singleshot_two_lb_zero_max.lp")),
        "files": ["unbounded_two_lb.lp"],
    },
    "unbounded_multishot_two_lb_one_max_superset": {
        "test": Supersets(get_model_from_file("unbounded_singleshot_two_lb_one_max.lp")),
        "files": ["unbounded_two_lb.lp"],
    },
    "unbounded_multishot_optimize_superset": {
        "test": Supersets(get_model_from_file("unbounded_singleshot_optimize.lp")),
        "files": ["unbounded_optimize.lp"],
    },
    "unbounded_multishot_optimize": {
        "test": StableModels(
            get_model_from_file("unbounded_singleshot_optimize.lp")
            | {
                'inc_set("root.bags")',
                'incremental("function","count(root.bags)",("count(root.bags)","count","root.bags"))',
                'incremental("minimize","count(root.bags)",("count(root.bags)",0))',
            }
        ),
        "files": ["unbounded_optimize.lp"],
    },
    "unbounded_multishot_zero_lb_zero_max": {
        "test": StableModels(
            get_model_from_file("unbounded_singleshot_zero_lb_zero_max.lp")
            | {
                'inc_set("root.bags.size.volume")',
                'inc_set("root.bags.pockets")',
                'incremental("function","sum(root.bags.size.volume)",("sum(root.bags.size.volume)","sum","root.bags.size.volume"))',
                'incremental("function","count(root.bags.pockets)",("count(root.bags.pockets)","count","root.bags.pockets"))',
                'incremental("binary","root.totalVolume[0]=sum(root.bags.size.volume)",("root.totalVolume[0]=sum(root.bags.size.volume)","root.totalVolume[0]","=","sum(root.bags.size.volume)"))',
                'incremental("binary","5<count(root.bags.pockets)",("5<count(root.bags.pockets)","5","<","count(root.bags.pockets)"))',
                'incremental("constraint","root.totalVolume[0]=sum(root.bags.size.volume)",((0,"root.totalVolume[0]=sum(root.bags.size.volume)"),"boolean"))',
                'incremental("constraint","5<count(root.bags.pockets)",((4,"5<count(root.bags.pockets)"),"boolean"))',
            }
        ),
        "files": ["unbounded_zero_lb.lp"],
    },
    "unbounded_multishot_two_lb_zero_max": {
        "test": StableModels(
            get_model_from_file("unbounded_singleshot_two_lb_zero_max.lp")
            | {
                'inc_set("root.bags.size.volume")',
                'inc_set("root.bags.pockets")',
                'inc_set("root.bags[0].pockets")',
                'inc_set("root.bags[1].pockets")',
                'incremental("function","sum(root.bags.size.volume)",("sum(root.bags.size.volume)","sum","root.bags.size.volume"))',
                'incremental("function","count(root.bags.pockets)",("count(root.bags.pockets)","count","root.bags.pockets"))',
                'incremental("function","count(root.bags[0].pockets)",("count(root.bags[0].pockets)","count","root.bags[0].pockets"))',
                'incremental("function","count(root.bags[1].pockets)",("count(root.bags[1].pockets)","count","root.bags[1].pockets"))',
                'incremental("constraint","root.bags[0]",((7,"root.bags[0]"),"table"))',
                'incremental("constraint","root.bags[1]",((7,"root.bags[1]"),"table"))',
                'incremental("binary","root.totalVolume[0]=sum(root.bags.size.volume)",("root.totalVolume[0]=sum(root.bags.size.volume)","root.totalVolume[0]","=","sum(root.bags.size.volume)"))',
                'incremental("binary","5<count(root.bags.pockets)",("5<count(root.bags.pockets)","5","<","count(root.bags.pockets)"))',
                'incremental("binary","2<=count(root.bags[0].pockets)",("2<=count(root.bags[0].pockets)","2","<=","count(root.bags[0].pockets)"))',
                'incremental("binary","2<=count(root.bags[1].pockets)",("2<=count(root.bags[1].pockets)","2","<=","count(root.bags[1].pockets)"))',
                'incremental("constraint","root.totalVolume[0]=sum(root.bags.size.volume)",((0,"root.totalVolume[0]=sum(root.bags.size.volume)"),"boolean"))',
                'incremental("constraint","5<count(root.bags.pockets)",((4,"5<count(root.bags.pockets)"),"boolean"))',
                'incremental("binary","!count(root.bags[0])=1||2<=count(root.bags[0].pockets)",("!count(root.bags[0])=1||2<=count(root.bags[0].pockets)","!count(root.bags[0])=1","||","2<=count(root.bags[0].pockets)"))',
                'incremental("binary","!count(root.bags[1])=1||2<=count(root.bags[1].pockets)",("!count(root.bags[1])=1||2<=count(root.bags[1].pockets)","!count(root.bags[1])=1","||","2<=count(root.bags[1].pockets)"))',
                'incremental("constraint","!count(root.bags[0])=1||2<=count(root.bags[0].pockets)",((5,"!count(root.bags[0])=1||2<=count(root.bags[0].pockets)"),"boolean"))',
                'incremental("constraint","!count(root.bags[1])=1||2<=count(root.bags[1].pockets)",((5,"!count(root.bags[1])=1||2<=count(root.bags[1].pockets)"),"boolean"))',
            }
        ),
        "files": ["unbounded_two_lb.lp"],
    },
    "unbounded_multishot_zero_lb_one_max": {
        "test": StableModels(
            get_model_from_file("unbounded_singleshot_zero_lb_one_max.lp")
            | {
                'inc_set("root.bags.size.volume")',
                'inc_set("root.bags.pockets")',
                'inc_set("root.bags[0].pockets")',
                'incremental("function","sum(root.bags.size.volume)",("sum(root.bags.size.volume)","sum","root.bags.size.volume"))',
                'incremental("function","count(root.bags.pockets)",("count(root.bags.pockets)","count","root.bags.pockets"))',
                'incremental("function","count(root.bags[0].pockets)",("count(root.bags[0].pockets)","count","root.bags[0].pockets"))',
                'incremental("constraint","root.bags[0]",((7,"root.bags[0]"),"table"))',
                'incremental("binary","root.totalVolume[0]=sum(root.bags.size.volume)",("root.totalVolume[0]=sum(root.bags.size.volume)","root.totalVolume[0]","=","sum(root.bags.size.volume)"))',
                'incremental("binary","5<count(root.bags.pockets)",("5<count(root.bags.pockets)","5","<","count(root.bags.pockets)"))',
                'incremental("binary","2<=count(root.bags[0].pockets)",("2<=count(root.bags[0].pockets)","2","<=","count(root.bags[0].pockets)"))',
                'incremental("constraint","root.totalVolume[0]=sum(root.bags.size.volume)",((0,"root.totalVolume[0]=sum(root.bags.size.volume)"),"boolean"))',
                'incremental("constraint","5<count(root.bags.pockets)",((4,"5<count(root.bags.pockets)"),"boolean"))',
                'incremental("binary","!count(root.bags[0])=1||2<=count(root.bags[0].pockets)",("!count(root.bags[0])=1||2<=count(root.bags[0].pockets)","!count(root.bags[0])=1","||","2<=count(root.bags[0].pockets)"))',
                'incremental("constraint","!count(root.bags[0])=1||2<=count(root.bags[0].pockets)",((5,"!count(root.bags[0])=1||2<=count(root.bags[0].pockets)"),"boolean"))',
            }
        ),
        "files": ["unbounded_zero_lb.lp"],
    },
    "unbounded_multishot_two_lb_one_max": {
        "test": StableModels(
            get_model_from_file("unbounded_singleshot_two_lb_one_max.lp")
            | {
                'inc_set("root.bags.size.volume")',
                'inc_set("root.bags.pockets")',
                'inc_set("root.bags[0].pockets")',
                'inc_set("root.bags[1].pockets")',
                'inc_set("root.bags[2].pockets")',
                'incremental("function","sum(root.bags.size.volume)",("sum(root.bags.size.volume)","sum","root.bags.size.volume"))',
                'incremental("function","count(root.bags.pockets)",("count(root.bags.pockets)","count","root.bags.pockets"))',
                'incremental("function","count(root.bags[0].pockets)",("count(root.bags[0].pockets)","count","root.bags[0].pockets"))',
                'incremental("function","count(root.bags[1].pockets)",("count(root.bags[1].pockets)","count","root.bags[1].pockets"))',
                'incremental("function","count(root.bags[2].pockets)",("count(root.bags[2].pockets)","count","root.bags[2].pockets"))',
                'incremental("constraint","root.bags[0]",((7,"root.bags[0]"),"table"))',
                'incremental("constraint","root.bags[1]",((7,"root.bags[1]"),"table"))',
                'incremental("constraint","root.bags[2]",((7,"root.bags[2]"),"table"))',
                'incremental("binary","root.totalVolume[0]=sum(root.bags.size.volume)",("root.totalVolume[0]=sum(root.bags.size.volume)","root.totalVolume[0]","=","sum(root.bags.size.volume)"))',
                'incremental("binary","5<count(root.bags.pockets)",("5<count(root.bags.pockets)","5","<","count(root.bags.pockets)"))',
                'incremental("binary","2<=count(root.bags[0].pockets)",("2<=count(root.bags[0].pockets)","2","<=","count(root.bags[0].pockets)"))',
                'incremental("binary","2<=count(root.bags[1].pockets)",("2<=count(root.bags[1].pockets)","2","<=","count(root.bags[1].pockets)"))',
                'incremental("binary","2<=count(root.bags[2].pockets)",("2<=count(root.bags[2].pockets)","2","<=","count(root.bags[2].pockets)"))',
                'incremental("constraint","root.totalVolume[0]=sum(root.bags.size.volume)",((0,"root.totalVolume[0]=sum(root.bags.size.volume)"),"boolean"))',
                'incremental("constraint","5<count(root.bags.pockets)",((4,"5<count(root.bags.pockets)"),"boolean"))',
                'incremental("binary","!count(root.bags[0])=1||2<=count(root.bags[0].pockets)",("!count(root.bags[0])=1||2<=count(root.bags[0].pockets)","!count(root.bags[0])=1","||","2<=count(root.bags[0].pockets)"))',
                'incremental("binary","!count(root.bags[1])=1||2<=count(root.bags[1].pockets)",("!count(root.bags[1])=1||2<=count(root.bags[1].pockets)","!count(root.bags[1])=1","||","2<=count(root.bags[1].pockets)"))',
                'incremental("binary","!count(root.bags[2])=1||2<=count(root.bags[2].pockets)",("!count(root.bags[2])=1||2<=count(root.bags[2].pockets)","!count(root.bags[2])=1","||","2<=count(root.bags[2].pockets)"))',
                'incremental("constraint","!count(root.bags[0])=1||2<=count(root.bags[0].pockets)",((5,"!count(root.bags[0])=1||2<=count(root.bags[0].pockets)"),"boolean"))',
                'incremental("constraint","!count(root.bags[1])=1||2<=count(root.bags[1].pockets)",((5,"!count(root.bags[1])=1||2<=count(root.bags[1].pockets)"),"boolean"))',
                'incremental("constraint","!count(root.bags[2])=1||2<=count(root.bags[2].pockets)",((5,"!count(root.bags[2])=1||2<=count(root.bags[2].pockets)"),"boolean"))',
            }
        ),
        "files": ["unbounded_two_lb.lp"],
    },
}



"""
Test cases for main application functionality.
"""

from io import StringIO
from unittest import TestCase

from clingo.symbol import parse_term

from coomsuite.utils import asp2coom, coom2asp, logging
from coomsuite.utils.logging import configure_logging, get_logger
from coomsuite.utils.parser import get_parser


class TestMain(TestCase):
    """
    Test cases for main application functionality.
    """

    def test_logger(self) -> None:
        """
        Test the logger.
        """
        sio = StringIO()
        configure_logging(sio, logging.INFO, True)
        log = get_logger("main")
        log.info("test123")
        self.assertRegex(sio.getvalue(), "test123")

    def test_parser(self) -> None:
        """
        Test the parser.
        """
        parser = get_parser()
        ret = parser.parse_args(["--log", "info"])
        self.assertEqual(ret.log, logging.INFO)

    def test_asp2coom(self) -> None:
        """
        Test the COOM output formatting.
        """
        self.assertEqual(asp2coom(parse_term('include("root.carrier[0]")')), "carrier[0]")
        self.assertEqual(asp2coom(parse_term('value("root.color[0]", "Blue")')), 'color[0] = "Blue"')
        self.assertEqual(asp2coom(parse_term('value("root.wheel[0].size[0]", 27)')), "wheel[0].size[0] = 27")
        self.assertRaises(ValueError, asp2coom, parse_term('instance("")'))

    def test_coom2asp(self) -> None:
        """
        Test the COOM to ASP conversion
        """
        self.assertEqual(coom2asp("carrier[0]"), 'include("root.carrier[0]")')
        self.assertEqual(coom2asp('color[0] = "Blue"'), 'value("root.color[0]","Blue")')
        self.assertEqual(coom2asp("wheel[0].size[0] = 27"), 'value("root.wheel[0].size[0]",27)')



# Generated from /home/heuer/tools/coom-suite/src/coomsuite/utils/coom_grammar/user/UserInput.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .UserInputParser import UserInputParser
else:
    from UserInputParser import UserInputParser

# This class defines a complete listener for a parse tree produced by UserInputParser.
class UserInputListener(ParseTreeListener):

    # Enter a parse tree produced by UserInputParser#user_input.
    def enterUser_input(self, ctx:UserInputParser.User_inputContext):
        pass

    # Exit a parse tree produced by UserInputParser#user_input.
    def exitUser_input(self, ctx:UserInputParser.User_inputContext):
        pass


    # Enter a parse tree produced by UserInputParser#input_block.
    def enterInput_block(self, ctx:UserInputParser.Input_blockContext):
        pass

    # Exit a parse tree produced by UserInputParser#input_block.
    def exitInput_block(self, ctx:UserInputParser.Input_blockContext):
        pass


    # Enter a parse tree produced by UserInputParser#input_operation.
    def enterInput_operation(self, ctx:UserInputParser.Input_operationContext):
        pass

    # Exit a parse tree produced by UserInputParser#input_operation.
    def exitInput_operation(self, ctx:UserInputParser.Input_operationContext):
        pass


    # Enter a parse tree produced by UserInputParser#set_value.
    def enterSet_value(self, ctx:UserInputParser.Set_valueContext):
        pass

    # Exit a parse tree produced by UserInputParser#set_value.
    def exitSet_value(self, ctx:UserInputParser.Set_valueContext):
        pass


    # Enter a parse tree produced by UserInputParser#add_instance.
    def enterAdd_instance(self, ctx:UserInputParser.Add_instanceContext):
        pass

    # Exit a parse tree produced by UserInputParser#add_instance.
    def exitAdd_instance(self, ctx:UserInputParser.Add_instanceContext):
        pass


    # Enter a parse tree produced by UserInputParser#floating.
    def enterFloating(self, ctx:UserInputParser.FloatingContext):
        pass

    # Exit a parse tree produced by UserInputParser#floating.
    def exitFloating(self, ctx:UserInputParser.FloatingContext):
        pass


    # Enter a parse tree produced by UserInputParser#path.
    def enterPath(self, ctx:UserInputParser.PathContext):
        pass

    # Exit a parse tree produced by UserInputParser#path.
    def exitPath(self, ctx:UserInputParser.PathContext):
        pass


    # Enter a parse tree produced by UserInputParser#path_item.
    def enterPath_item(self, ctx:UserInputParser.Path_itemContext):
        pass

    # Exit a parse tree produced by UserInputParser#path_item.
    def exitPath_item(self, ctx:UserInputParser.Path_itemContext):
        pass


    # Enter a parse tree produced by UserInputParser#path_index.
    def enterPath_index(self, ctx:UserInputParser.Path_indexContext):
        pass

    # Exit a parse tree produced by UserInputParser#path_index.
    def exitPath_index(self, ctx:UserInputParser.Path_indexContext):
        pass


    # Enter a parse tree produced by UserInputParser#name.
    def enterName(self, ctx:UserInputParser.NameContext):
        pass

    # Exit a parse tree produced by UserInputParser#name.
    def exitName(self, ctx:UserInputParser.NameContext):
        pass


    # Enter a parse tree produced by UserInputParser#stmt_end.
    def enterStmt_end(self, ctx:UserInputParser.Stmt_endContext):
        pass

    # Exit a parse tree produced by UserInputParser#stmt_end.
    def exitStmt_end(self, ctx:UserInputParser.Stmt_endContext):
        pass


    # Enter a parse tree produced by UserInputParser#formula_atom.
    def enterFormula_atom(self, ctx:UserInputParser.Formula_atomContext):
        pass

    # Exit a parse tree produced by UserInputParser#formula_atom.
    def exitFormula_atom(self, ctx:UserInputParser.Formula_atomContext):
        pass



del UserInputParser



"""
Contains custom clintest Tests
"""

from typing import Set, Union

from clingo import Symbol
from clingo.solving import Model
from clintest.assertion import Equals, False_, Implies, Optimal, SubsetOf, SupersetOf, True_
from clintest.quantifier import All, Exact, Last
from clintest.test import And, Assert, Test

TEST_EMPTY = Assert(All(), SubsetOf(set()))
TEST_UNSAT = Assert(Exact(0), False_())


def NumModels(n: int) -> Test:  # pylint: disable=invalid-name
    """
    clintest.Test for checking that a program has a certain number of models
    """
    return Assert(Exact(n), True_())


def StableModels(*args: set[Symbol | str]) -> Test:  # pylint: disable=invalid-name
    """
    clintest.Test for checking that a program has a certain set of stable models.
    Works with clingo and flingo

    Args:
        args: The set of stable models as a set of sets of strings or clingo symbols
    """
    return And(NumModels(len(args)), *(Assert(Exact(1), EqualsFlingo(a)) for a in args))


def OptimalModel(model: set[Symbol | str], flingo: bool = False) -> Test:  # pylint: disable=invalid-name
    """
    clintest.Test for checking that a program has a certain optimal stable model

    Args:
        args: The optimal stable model as a set of sets of strings or clingo symbols
        flingo (bool): Whether to prepare the test for use with the flingo solver
    """
    if flingo:
        return Assert(Last(), EqualsFlingo(model))  # No optimality statements in flingo
    return Assert(All(), Implies(Optimal(), Equals(model)))


def Supersets(*args: set[Symbol | str]) -> Test:  # pylint: disable=invalid-name
    """
    clintest.Test for checking models of a program against a set of models using superset relation
    """
    return And(NumModels(len(args)), *(Assert(Exact(1), SupersetOf(a)) for a in args))


class EqualsFlingo(Equals):
    """
    A clintest Equals assertion that can also check theory atoms and removes auxiliary flingo atoms.

    Args:
        symbol (Symbol): A clingo symbol.
    """

    def __init__(self, symbols: Set[Union[Symbol, str]]) -> None:
        super().__init__(symbols)
        self.__symbols = self._Equals__symbols  # type: ignore # pylint: disable=no-member

    def holds_for(self, model: Model) -> bool:
        return self.__symbols == {
            s for s in model.symbols(shown=True, theory=True) if s.name not in ["__csp", "__def", "__csp_cost"]
        }


# class SupersetOfTheory(SupersetOf):
#     """
#     A clintest SupersetOf assertion that can also check theory atoms.

#     Args:
#         symbol (Symbol): A clingo symbol.
#         check_theory (bool): Whether to include theory atoms in the check
#     """

#     def __init__(self, symbols: Set[Union[Symbol, str]], check_theory: bool = False) -> None:
#         super().__init__(symbols)
#         self.__symbols = self._SupersetOf__symbols  # type: ignore # pylint: disable=no-member
#         self.__check_theory = check_theory

#     def holds_for(self, model: Model) -> bool:
#         if self.__check_theory:
#             return set(model.symbols(shown=True, theory=True)).issuperset(self.__symbols)
#         return super().holds_for(model)  # nocoverage

# class ContainsTheory(Contains):
#     """
#     A clintest Contains assertion that can also check theory atoms.

#     Args:
#         symbol (Symbol): A clingo symbol.
#         check_theory (bool): Whether to include theory atoms in the check
#     """

#     def __init__(self, symbol: Union[Symbol, str], check_theory: bool = False) -> None:
#         super().__init__(symbol)
#         self.__symbol = self._Contains__symbol  # type: ignore # pylint: disable=no-member
#         self.__check_theory = check_theory

#     def holds_for(self, model: Model) -> bool:
#         if self.__check_theory:
#             return self.__symbol in model.symbols(shown=True, theory=True)
#         return super().holds_for(model)  # nocoverage



# Generated from /home/heuer/tools/coom-suite/src/coomsuite/utils/coom_grammar/user/UserInput.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .UserInputParser import UserInputParser
else:
    from UserInputParser import UserInputParser

# This class defines a complete generic visitor for a parse tree produced by UserInputParser.

class UserInputVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by UserInputParser#user_input.
    def visitUser_input(self, ctx:UserInputParser.User_inputContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by UserInputParser#input_block.
    def visitInput_block(self, ctx:UserInputParser.Input_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by UserInputParser#input_operation.
    def visitInput_operation(self, ctx:UserInputParser.Input_operationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by UserInputParser#set_value.
    def visitSet_value(self, ctx:UserInputParser.Set_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by UserInputParser#add_instance.
    def visitAdd_instance(self, ctx:UserInputParser.Add_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by UserInputParser#floating.
    def visitFloating(self, ctx:UserInputParser.FloatingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by UserInputParser#path.
    def visitPath(self, ctx:UserInputParser.PathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by UserInputParser#path_item.
    def visitPath_item(self, ctx:UserInputParser.Path_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by UserInputParser#path_index.
    def visitPath_index(self, ctx:UserInputParser.Path_indexContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by UserInputParser#name.
    def visitName(self, ctx:UserInputParser.NameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by UserInputParser#stmt_end.
    def visitStmt_end(self, ctx:UserInputParser.Stmt_endContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by UserInputParser#formula_atom.
    def visitFormula_atom(self, ctx:UserInputParser.Formula_atomContext):
        return self.visitChildren(ctx)



del UserInputParser



"""
Test cases for the parser from COOM to ASP.
"""

# pylint: disable=too-many-lines

from unittest import TestCase

from . import parse_coom


class TestCOOMUserParser(TestCase):
    """
    Test cases for COOM User input to ASP parser.
    """

    def test_set(self) -> None:
        """
        Test parsing the 'set' keyword
        used for setting values.
        """
        self.assertEqual(parse_coom("set color[0] = Red", grammar="user"), ['user_value("root.color[0]","Red").'])
        self.assertEqual(
            parse_coom("set frontWheel[0].size[0] = 23", grammar="user"),
            ['user_value("root.frontWheel[0].size[0]",23).'],
        )

    def test_add(self) -> None:
        """
        Test parsing the 'add' keyword
        used for adding instances of objects.
        """
        self.assertEqual(parse_coom("add basket[0]", grammar="user"), ['user_include("root.basket[0]").'])
        self.assertEqual(
            parse_coom("add wheels[0] add wheels[1]", grammar="user"),
            ['user_include("root.wheels[0]").', 'user_include("root.wheels[1]").'],
        )

    def test_blockinput(self) -> None:
        """
        Test parsing a user input block.
        """
        self.assertEqual(
            parse_coom("blockinput basket[0] {set color[0] = Yellow}", grammar="user"),
            ['user_value("root.basket[0].color[0]","Yellow").'],
        )


class TestCOOMModelParser(TestCase):
    """
    Test cases for COOM Model to ASP parser.
    """

    def test_product(self) -> None:
        """
        Test parsing the 'product' keyword.
        """
        self.assertEqual(parse_coom("product{}"), ['structure("product").'])

    def test_structure(self) -> None:
        """
        Test parsing the 'structure' keyword.
        """
        self.assertEqual(parse_coom("structure Test {}"), ['structure("Test").'])

    def test_feature(self) -> None:
        """
        Test parsing the COOM features.
        """
        self.assertEqual(
            parse_coom("product{Wheel wheel Frame frame}"),
            [
                'structure("product").',
                'feature("product","wheel","Wheel",1,1).',
                'feature("product","frame","Frame",1,1).',
            ],
        )
        self.assertEqual(
            parse_coom("structure Carrier {0..3 Bag bag}"),
            ['structure("Carrier").', 'feature("Carrier","bag","Bag",0,3).'],
        )
        self.assertEqual(
            parse_coom("structure Carrier {2..* Bag bag}"),
            ['structure("Carrier").', 'feature("Carrier","bag","Bag",2,#sup).'],
        )
        self.assertEqual(
            parse_coom("structure Carrier {5x Bag bag}"),
            ['structure("Carrier").', 'feature("Carrier","bag","Bag",5,5).'],
        )
        self.assertEqual(
            parse_coom("product{num	1-100 totalWeight}"),
            [
                'structure("product").',
                'feature("product","totalWeight","num",1,1).',
                'range("product","totalWeight",1,100).',
            ],
        )
        self.assertEqual(
            parse_coom("product{num	-100-10 totalWeight}"),
            [
                'structure("product").',
                'feature("product","totalWeight","num",1,1).',
                'range("product","totalWeight",-100,10).',
            ],
        )
        self.assertEqual(
            parse_coom("product{num 0-∞ totalWeight}"),
            [
                'structure("product").',
                'feature("product","totalWeight","num",1,1).',
                'range("product","totalWeight",0,#sup).',
            ],
        )
        self.assertEqual(
            parse_coom("product{num -∞-1000 totalWeight}"),
            [
                'structure("product").',
                'feature("product","totalWeight","num",1,1).',
                'range("product","totalWeight",#inf,1000).',
            ],
        )

        self.assertEqual(parse_coom("product{string id}"), ['structure("product").'])

    def test_enumeration(self) -> None:
        """
        Test parsing the 'enumeration' keyword.
        """
        self.assertEqual(
            parse_coom("enumeration Color {Red Green Yellow Blue}"),
            [
                'enumeration("Color").',
                'option("Color", "Red").',
                'option("Color", "Green").',
                'option("Color", "Yellow").',
                'option("Color", "Blue").',
            ],
        )
        self.assertEqual(
            parse_coom(
                """\
                enumeration Capacity {
                        attribute num/l volume
                        attribute num/gr weight
                        B10  = ( 10,  100)
                        B20  = ( 20,  250)
                    }"""
            ),
            [
                'enumeration("Capacity").',
                'attribute("Capacity","volume","num").',
                'attribute("Capacity","weight","num").',
                'option("Capacity", "B10").',
                'attribute_value("Capacity","B10","volume",10).',
                'attribute_value("Capacity","B10","weight",100).',
                'option("Capacity", "B20").',
                'attribute_value("Capacity","B20","volume",20).',
                'attribute_value("Capacity","B20","weight",250).',
            ],
        )

        self.assertEqual(
            parse_coom(
                """\
                enumeration Size {
                        attribute color
                        Small  = ( Red )
                        Big    = ( Blue )
                    }"""
            ),
            [
                'enumeration("Size").',
                'attribute("Size","color","str").',
                'option("Size", "Small").',
                'attribute_value("Size","Small","color","Red").',
                'option("Size", "Big").',
                'attribute_value("Size","Big","color","Blue").',
            ],
        )

    def test_behavior(self) -> None:
        """
        Test parsing the 'behavior' keyword.
        """
        self.assertEqual(
            parse_coom("behavior{require a = b}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=b").',
                'binary("a=b","a","=","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior Bag {require a = b}"),
            [
                "behavior(0).",
                'context(0,"Bag").',
                'require(0,"a=b").',
                'binary("a=b","a","=","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior {require color = Red}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"color=Red").',
                'binary("color=Red","color","=","Red").',
                'path("color",0,"color").',
                'constant("Red").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior {default color = red}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'default(0,"color","red").',
                'path("color",0,"color").',
                'path("red",0,"red").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = b require c = d}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=b").',
                'binary("a=b","a","=","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
                "behavior(1).",
                'context(1,"product").',
                'require(1,"c=d").',
                'binary("c=d","c","=","d").',
                'path("c",0,"c").',
                'path("d",0,"d").',
            ],
        )

        self.assertEqual(
            parse_coom(
                """behavior {
                            combinations  (wheelSupport	 rearWheel)
                            allow         (True          (W14, W16))
                            allow         (False         (W18, W20))}"""
            ),
            [
                "behavior(0).",
                'context(0,"product").',
                'combinations(0,0,"wheelSupport").',
                'combinations(0,1,"rearWheel").',
                'path("wheelSupport",0,"wheelSupport").',
                'path("rearWheel",0,"rearWheel").',
                'allow(0,(0,0),"True").',
                'allow(0,(1,0),"W14").',
                'allow(0,(1,0),"W16").',
                'allow(0,(0,1),"False").',
                'allow(0,(1,1),"W18").',
                'allow(0,(1,1),"W20").',
            ],
        )

        self.assertEqual(
            parse_coom(
                """behavior {
                            combinations  (color	 size)
                            allow         (Red       -*-)
                            allow         (Yellow    S)}"""
            ),
            [
                "behavior(0).",
                'context(0,"product").',
                'combinations(0,0,"color").',
                'combinations(0,1,"size").',
                'path("color",0,"color").',
                'path("size",0,"size").',
                'allow(0,(0,0),"Red").',
                'allow(0,(0,1),"Yellow").',
                'allow(0,(1,1),"S").',
            ],
        )

        self.assertEqual(
            parse_coom(
                """behavior Bike {
                            combinations  (wheelSupport	 rearWheel)
                            allow         (True          (W14, W16))
                            allow         (False         (W18, W20))}"""
            ),
            [
                "behavior(0).",
                'context(0,"Bike").',
                'combinations(0,0,"wheelSupport").',
                'combinations(0,1,"rearWheel").',
                'path("wheelSupport",0,"wheelSupport").',
                'path("rearWheel",0,"rearWheel").',
                'allow(0,(0,0),"True").',
                'allow(0,(1,0),"W14").',
                'allow(0,(1,0),"W16").',
                'allow(0,(0,1),"False").',
                'allow(0,(1,1),"W18").',
                'allow(0,(1,1),"W20").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior {condition a = b require c > 5}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'condition(0,0,"a=b").',
                'binary("a=b","a","=","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
                'require(0,"c>5").',
                'binary("c>5","c",">","5").',
                'path("c",0,"c").',
                'number("5",5).',
            ],
        )

        self.assertEqual(
            parse_coom("behavior {condition a = 2 condition b = 1 require c > 5}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'condition(0,0,"a=2").',
                'binary("a=2","a","=","2").',
                'path("a",0,"a").',
                'number("2",2).',
                'condition(0,1,"b=1").',
                'binary("b=1","b","=","1").',
                'path("b",0,"b").',
                'number("1",1).',
                'require(0,"c>5").',
                'binary("c>5","c",">","5").',
                'path("c",0,"c").',
                'number("5",5).',
            ],
        )

        self.assertEqual(
            parse_coom("behavior {condition a = b default c = d}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'condition(0,0,"a=b").',
                'binary("a=b","a","=","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
                'default(0,"c","d").',
                'path("c",0,"c").',
                'path("d",0,"d").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{imply a = b}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'imply(0,"a","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(parse_coom("behavior{readonly totalWeight}"), [])

    def test_formula(self) -> None:
        """
        Test parsing with Boolean and arithmetic formulas.
        """
        self.assertEqual(
            parse_coom("behavior{require a = b || a = c}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=b||a=c").',
                'binary("a=b||a=c","a=b","||","a=c").',
                'binary("a=b","a","=","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
                'binary("a=c","a","=","c").',
                'path("a",0,"a").',
                'path("c",0,"c").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = b || a = c || a = d}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=b||a=c||a=d").',
                'binary("a=b||a=c||a=d","a=b","||","a=c||a=d").',
                'binary("a=c||a=d","a=c","||","a=d").',
                'binary("a=b","a","=","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
                'binary("a=c","a","=","c").',
                'path("a",0,"a").',
                'path("c",0,"c").',
                'binary("a=d","a","=","d").',
                'path("a",0,"a").',
                'path("d",0,"d").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = b && a = c}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=b&&a=c").',
                'binary("a=b&&a=c","a=b","&&","a=c").',
                'binary("a=b","a","=","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
                'binary("a=c","a","=","c").',
                'path("a",0,"a").',
                'path("c",0,"c").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = b && a = c && a = d}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=b&&a=c&&a=d").',
                'binary("a=b&&a=c&&a=d","a=b","&&","a=c&&a=d").',
                'binary("a=c&&a=d","a=c","&&","a=d").',
                'binary("a=b","a","=","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
                'binary("a=c","a","=","c").',
                'path("a",0,"a").',
                'path("c",0,"c").',
                'binary("a=d","a","=","d").',
                'path("a",0,"a").',
                'path("d",0,"d").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require ! a = b }"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"!a=b").',
                'unary("!a=b","!","a=b").',
                'binary("a=b","a","=","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require (a = b) }"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"(a=b)").',
                'unary("(a=b)","()","a=b").',
                'binary("a=b","a","=","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = b + c}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=b+c").',
                'binary("a=b+c","a","=","b+c").',
                'path("a",0,"a").',
                'binary("b+c","b","+","c").',
                'path("b",0,"b").',
                'path("c",0,"c").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = b - c}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=b-c").',
                'binary("a=b-c","a","=","b-c").',
                'path("a",0,"a").',
                'binary("b-c","b","-","c").',
                'path("b",0,"b").',
                'path("c",0,"c").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = b * c}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=b*c").',
                'binary("a=b*c","a","=","b*c").',
                'path("a",0,"a").',
                'binary("b*c","b","*","c").',
                'path("b",0,"b").',
                'path("c",0,"c").',
            ],
        )
        self.assertEqual(
            parse_coom("behavior{require a = b / c}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=b/c").',
                'binary("a=b/c","a","=","b/c").',
                'path("a",0,"a").',
                'binary("b/c","b","/","c").',
                'path("b",0,"b").',
                'path("c",0,"c").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = b ^ c}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=b^c").',
                'binary("a=b^c","a","=","b^c").',
                'path("a",0,"a").',
                'binary("b^c","b","^","c").',
                'path("b",0,"b").',
                'path("c",0,"c").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = - b}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=-b").',
                'binary("a=-b","a","=","-b").',
                'path("a",0,"a").',
                'unary("-b","-","b").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = + b}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=+b").',
                'binary("a=+b","a","=","+b").',
                'path("a",0,"a").',
                'unary("+b","+","b").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = (b)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=(b)").',
                'binary("a=(b)","a","=","(b)").',
                'path("a",0,"a").',
                'unary("(b)","()","b").',
                'path("b",0,"b").',
            ],
        )

    def test_function(self) -> None:
        """
        Test parsing behavior with functions.
        """

        self.assertEqual(
            parse_coom("behavior{require a = sum(b)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=sum(b)").',
                'binary("a=sum(b)","a","=","sum(b)").',
                'path("a",0,"a").',
                'function("product","sum(b)","sum","b").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = sum(b,c)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=sum(b,c)").',
                'binary("a=sum(b,c)","a","=","sum(b,c)").',
                'path("a",0,"a").',
                'function("product","sum(b,c)","sum","b").',
                'function("product","sum(b,c)","sum","c").',
                'path("b",0,"b").',
                'path("c",0,"c").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = count(b)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=count(b)").',
                'binary("a=count(b)","a","=","count(b)").',
                'path("a",0,"a").',
                'function("product","count(b)","count","b").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = count(b,c)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=count(b,c)").',
                'binary("a=count(b,c)","a","=","count(b,c)").',
                'path("a",0,"a").',
                'function("product","count(b,c)","count","b").',
                'function("product","count(b,c)","count","c").',
                'path("b",0,"b").',
                'path("c",0,"c").',
            ],
        )
        self.assertEqual(
            parse_coom("behavior{require a = min(b)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=min(b)").',
                'binary("a=min(b)","a","=","min(b)").',
                'path("a",0,"a").',
                'function("product","min(b)","min","b").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = min(b,c)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=min(b,c)").',
                'binary("a=min(b,c)","a","=","min(b,c)").',
                'path("a",0,"a").',
                'function("product","min(b,c)","min","b").',
                'function("product","min(b,c)","min","c").',
                'path("b",0,"b").',
                'path("c",0,"c").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = max(b)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=max(b)").',
                'binary("a=max(b)","a","=","max(b)").',
                'path("a",0,"a").',
                'function("product","max(b)","max","b").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = max(b,c)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=max(b,c)").',
                'binary("a=max(b,c)","a","=","max(b,c)").',
                'path("a",0,"a").',
                'function("product","max(b,c)","max","b").',
                'function("product","max(b,c)","max","c").',
                'path("b",0,"b").',
                'path("c",0,"c").',
            ],
        )

        # self.assertEqual(
        #     parse_coom("behavior{require a = delta(b)}"),
        #     [
        #         "behavior(0).",
        #         'context(0,"product").',
        #         'require(0,"a=delta(b)").',
        #         'binary("a=delta(b)","a","=","delta(b)").',
        #         'path("a",0,"a").',
        #         'unary("product","delta(b)","delta","b").',
        #         'path("b",0,"b").',
        #     ],
        # )

        # self.assertEqual(
        #     parse_coom("behavior{require a = delta(b,c)}"),
        #     [
        #         "behavior(0).",
        #         'context(0,"product").',
        #         'require(0,"a=delta(b,c)").',
        #         'binary("a=delta(b,c)","a","=","delta(b,c)").',
        #         'path("a",0,"a").',
        #         'function("product","delta(b,c)","delta","b").',
        #         'function("product","delta(b,c)","delta","c").',
        #         'path("b",0,"b").',
        #         'path("c",0,"c").',
        #     ],
        # )

        # self.assertEqual(
        #     parse_coom("behavior{require a = pow(b)}"),
        #     [
        #         "behavior(0).",
        #         'context(0,"product").',
        #         'require(0,"a=pow(b)").',
        #         'binary("a=pow(b)","a","=","pow(b)").',
        #         'path("a",0,"a").',
        #         'function("product","pow(b)","pow","b").',
        #         'path("b",0,"b").',
        #     ],
        # )

        self.assertEqual(
            parse_coom("behavior{require a = sqrt(b)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=sqrt(b)").',
                'binary("a=sqrt(b)","a","=","sqrt(b)").',
                'path("a",0,"a").',
                'unary("sqrt(b)","sqrt","b").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = ceil(b)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=ceil(b)").',
                'binary("a=ceil(b)","a","=","ceil(b)").',
                'path("a",0,"a").',
                'unary("ceil(b)","ceil","b").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = floor(b)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=floor(b)").',
                'binary("a=floor(b)","a","=","floor(b)").',
                'path("a",0,"a").',
                'unary("floor(b)","floor","b").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = round(b)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=round(b)").',
                'binary("a=round(b)","a","=","round(b)").',
                'path("a",0,"a").',
                'unary("round(b)","round","b").',
                'path("b",0,"b").',
            ],
        )

        # self.assertEqual(
        #     parse_coom("behavior{require a = mod(b)}"),
        #     [
        #         "behavior(0).",
        #         'context(0,"product").',
        #         'require(0,"a=mod(b)").',
        #         'binary("a=mod(b)","a","=","mod(b)").',
        #         'path("a",0,"a").',
        #         'function("product","mod(b)","mod","b").',
        #         'path("b",0,"b").',
        #     ],
        # )

        self.assertEqual(
            parse_coom("behavior{require a = log(b)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=log(b)").',
                'binary("a=log(b)","a","=","log(b)").',
                'path("a",0,"a").',
                'unary("log(b)","log","b").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require a = ln(b)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=ln(b)").',
                'binary("a=ln(b)","a","=","ln(b)").',
                'path("a",0,"a").',
                'unary("ln(b)","ln","b").',
                'path("b",0,"b").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require x = sin(5)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"x=sin(5)").',
                'binary("x=sin(5)","x","=","sin(5)").',
                'path("x",0,"x").',
                'unary("sin(5)","sin","5").',
                'number("5",5).',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require x = asin(5)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"x=asin(5)").',
                'binary("x=asin(5)","x","=","asin(5)").',
                'path("x",0,"x").',
                'unary("asin(5)","asin","5").',
                'number("5",5).',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require x = cos(5)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"x=cos(5)").',
                'binary("x=cos(5)","x","=","cos(5)").',
                'path("x",0,"x").',
                'unary("cos(5)","cos","5").',
                'number("5",5).',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require x = acos(5)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"x=acos(5)").',
                'binary("x=acos(5)","x","=","acos(5)").',
                'path("x",0,"x").',
                'unary("acos(5)","acos","5").',
                'number("5",5).',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require x = cosh(5)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"x=cosh(5)").',
                'binary("x=cosh(5)","x","=","cosh(5)").',
                'path("x",0,"x").',
                'unary("cosh(5)","cosh","5").',
                'number("5",5).',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require x = tan(5)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"x=tan(5)").',
                'binary("x=tan(5)","x","=","tan(5)").',
                'path("x",0,"x").',
                'unary("tan(5)","tan","5").',
                'number("5",5).',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require x = atan(5)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"x=atan(5)").',
                'binary("x=atan(5)","x","=","atan(5)").',
                'path("x",0,"x").',
                'unary("atan(5)","atan","5").',
                'number("5",5).',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{require x = tanh(5)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"x=tanh(5)").',
                'binary("x=tanh(5)","x","=","tanh(5)").',
                'path("x",0,"x").',
                'unary("tanh(5)","tanh","5").',
                'number("5",5).',
            ],
        )

    def test_path(self) -> None:
        """
        Test parsing behavior with path expressions.
        """
        self.assertEqual(
            parse_coom("behavior{require a = b.c.d.e.f}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'require(0,"a=b.c.d.e.f").',
                'binary("a=b.c.d.e.f","a","=","b.c.d.e.f").',
                'path("a",0,"a").',
                'path("b.c.d.e.f",0,"b").',
                'path("b.c.d.e.f",1,"c").',
                'path("b.c.d.e.f",2,"d").',
                'path("b.c.d.e.f",3,"e").',
                'path("b.c.d.e.f",4,"f").',
            ],
        )

    def test_explanation(self) -> None:
        """
        Test parsing behavior with explanation.
        """
        self.assertEqual(
            parse_coom('behavior{explanation "This is a test constraint" require a = b}'),
            [
                "behavior(0).",
                'context(0,"product").',
                'explanation(0,"This is a test constraint").',
                'require(0,"a=b").',
                'binary("a=b","a","=","b").',
                'path("a",0,"a").',
                'path("b",0,"b").',
            ],
        )

    def test_optimization(self) -> None:
        """
        Test parsing behavior with optimization statements
        """

        self.assertEqual(
            parse_coom("behavior{minimize totalWeight}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'minimize(0,0,"totalWeight").',
                'path("totalWeight",0,"totalWeight").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{maximize totalOutput}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'maximize(0,0,"totalOutput").',
                'path("totalOutput",0,"totalOutput").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{minimize /3 totalWeight}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'minimize(0,3,"totalWeight").',
                'path("totalWeight",0,"totalWeight").',
            ],
        )

        self.assertEqual(
            parse_coom("behavior{maximize count(bags)}"),
            [
                "behavior(0).",
                'context(0,"product").',
                'maximize(0,0,"count(bags)").',
                'function("product","count(bags)","count","bags").',
                'path("bags",0,"bags").',
            ],
        )



"""
General helper functions for solving unbounded cardinalities.
"""

from itertools import chain, count, dropwhile
from typing import Iterator, Optional


def _exponential_iter(step: int) -> Iterator[int]:
    n = 0
    while True:
        yield step**n
        n += 1


def get_bound_iter(algorithm: str, start: int, step: int) -> Iterator[int]:
    """
    Get an iterator over the bounds for a selected algorithm.

    Note that the iterator starts after the start value.

    Args:
        algorithm (str): either "linear" or "exponential"
        start (int): value after which the iterator starts
        step (int): step size for linear/exponential base for iteration
    """
    iterator: Iterator[int]
    match algorithm:
        case "linear":
            iterator = count(start=start, step=step)
        case "exponential":
            iterator = chain([start], dropwhile(lambda x: x <= start, _exponential_iter(step)))
        case _:
            raise ValueError(f"unknown algorithm for bound iter: {algorithm}")

    return iterator


def next_bound_converge(unsat_bound: int, sat_bound: int) -> Optional[int]:
    """
    Determine the next bound (between unsat_bound and sat_bound) for converging to the minimal bound

    Returns:
        Optional[int]: The next bound to check, or None if sat_bound is already the minimal bound
    """
    if unsat_bound + 1 == sat_bound:
        return None

    return (unsat_bound + sat_bound) // 2



"""
The main entry point for the application.
"""

import sys
from os.path import basename, join, splitext
from tempfile import TemporaryDirectory

from . import convert_coom, solve, write_facts
from .bounds.solver import BoundSolver
from .preprocess import preprocess
from .utils.logging import configure_logging, get_logger
from .utils.parser import get_parser


def main() -> None:
    """
    Run the main function.
    """
    parser = get_parser()
    args, solver_args = parser.parse_known_args()
    configure_logging(sys.stderr, args.log, sys.stderr.isatty())

    log = get_logger("main")

    # log.info("info")
    # log.warning("warning")
    # log.debug("debug")
    # log.error("error")

    log.info("Converting COOM file %s", args.input)
    serialized_facts, unbounded, optimize = convert_coom(
        args.input, coom_user=args.user_input if args.user_input else None
    )

    if args.command == "convert":
        if args.output is None:
            print(serialized_facts)
        else:
            write_facts(serialized_facts, join(args.output, splitext(basename(args.input))[0] + "-coom.lp"))

    elif args.command == "solve":
        log.info("Solving COOM file %s", args.input)

        with TemporaryDirectory() as temp_dir:
            temp_file = join(temp_dir, "serialized-facts.lp")
            log.info("Saving serialized facts to %s", temp_file)
            write_facts(serialized_facts, temp_file)

            if args.show_facts:
                print(
                    "\n".join(
                        preprocess([temp_file], max_bound=args.initial_bound, discrete=True, multishot=args.multishot)
                    )
                )  # nocoverage
            elif unbounded:
                bound_solver = BoundSolver([temp_file], args.solver, solver_args + ["--opt-mode=ignore"], args.output)
                bound = bound_solver.get_bounds(
                    algorithm=args.bounds,
                    step=args.step,
                    initial_bound=args.initial_bound,
                    use_multishot=args.multishot,
                )
                if optimize and bound is not None:
                    print(f"\n Optimizing with minimal upper bound {bound}.")
                    solve([temp_file], args.solver, bound, solver_args, args.output)
                    print(f"\n The minimal upper bound is {bound}.")
                else:
                    print(f"\n The minimal upper bound is {bound}.")
            else:
                solve([temp_file], args.solver, 0, solver_args, args.output)


if __name__ == "__main__":
    main()
